# ### Server Operators ###
# If you want to customise any settings in this file, you'll need to work around
# the fact that is version controlled. Some options:
# 1. Use a local commit and always do a rebase pull on top of it. (This may
#    affect versioning, though.)
# 2. Create config.yml and write your overrides in there.
# 3. Use a more complicated scripting system; for example, dgamelaunch-config
#    supports templating this file and automatically handles merging the
#    server config with the repository files.
#
# ### Developers ###
# Warning! Servers will not update or merge with the version controlled copy of
# this file, so any parameters here should be presented as recommendations or
# documentation of the default, not a default value, and option name changes
# here will require manual intervention on the part of server admins. New
# values should always come with a default that does not lead to crashes, so
# that something sensible happens on DGL upgrade.
#
# To add new default values, see the `webtiles.config` module. Note that
# webtiles.config.get(x) returns `None` if `x` is not present in the config
# data, so this can be taken as a default default value.
#
# Bad assumptions about config defaults particularly impacts templated html
# files, which are loaded and called dynamically, so *do* get updated
# immediately on a rebuild (unlike python code in this module or
# webtiles.config). If something like client.html raises an exception, this
# will trigger 500 errors across the whole server.

import logging
import os

import yaml

# Where to look for `games.d/`, `config.yml`, and other things.
server_path = os.path.dirname(os.path.abspath(__file__))

# Set to False to enable `local webtiles` mode: the regular webtiles lobby
# is disabled, and opening webtiles loads the game main menu. (default: True)
# dgl_mode = False

# to enable ttyrec writing at the server level:
# enable_ttyrecs = False

# set `bind_nonsecure` to False to use only SSL. Set to `"redirect"` to bind
# the nonsecure ports as indicated, but redirect them to an SSL port.
bind_nonsecure = True
bind_address = ""
bind_port = 8080

# Or listen on multiple address/port pairs (overriding the above) with:
# bind_pairs = (
#     ("127.0.0.1", 8080),
#     ("localhost", 8082),
#     ("", 8180), # All addresses
# )

# set `ssl_options` to Tornado ssl options in order to use ssl on the specified
# port. For a real deployment, in addition to certfile/keyfile as below, you
# will probably need to provide "ca_certs" with your full CA cert chain.
# The server needs read access to keyfile when accepting new connections, so
# it should be readable by the user running the server (not just root).
#
# For local testing, you can generate a self-signed cert with a command like:
# ```
# openssl req -x509 -out localhost.crt -keyout localhost.key \
#   -newkey rsa:2048 -nodes -sha256 \
#   -subj '/CN=localhost' -extensions EXT -config <( \
#    printf "[dn]\nCN=localhost\n[req]\ndistinguished_name = dn\n[EXT]\nsubjectAltName=DNS:localhost\nkeyUsage=digitalSignature\nextendedKeyUsage=serverAuth")
# ```
# (this sort of cert will generate security warnings in modern browsers, so it
# is appropriate only for development use)

ssl_options = None # No SSL
# ssl_options = {
#    "certfile": "./webserver/localhost.crt",
#    "keyfile": "./webserver/localhost.key"
# }

ssl_address = ""
ssl_port = 8443

# Or listen on multiple address/port pairs (overriding the above) with:
# ssl_bind_pairs = (
#     ("127.0.0.1", 8081),
#     ("localhost", 8083),
# )



# to log to a file directly, uncomment and add something like:
#    "filename": "webtiles.log",
# logging_config = {
#     "level": logging.INFO,
#     "format": "%(asctime)s %(levelname)s: %(message)s"
# }

# sometimes useful for debugging:
# logging.getLogger('asyncio').setLevel(logging.DEBUG)

password_db = "./webserver/passwd.db3"
# Uncomment and change if you want this db somewhere separate from the
# password_db location.
#settings_db = "./webserver/user_settings.db3"

static_path = "./webserver/static"
template_path = "./webserver/templates/"

# Path for server-side unix sockets (to be used to communicate with crawl)
# server_socket_path = None # None: Uses global temp dir

# Server name, so far only used in the ttyrec metadata
server_id = ""

# Disable caching of game data files
game_data_no_cache = True

# Watch socket dirs for games not started by the server
# watch_socket_dirs = False

################
# Game configs
################
#
# The webtiles server needs some game modes configured in order to run. The
# default in-repo configuration enables regular mode, seeded, tutorial, and
# sprint for the `crawl` binary as built in the repository. (Build with
# `WEBTILES=y`.)
#
# You can define game configs in two ways:
# 1. As *.yml files in `games.d/`. (default; see games.d/base.yml)
# 2. With a dictionary `games` in this file
#
# Both of these formats are essentially the same. See the commented examples
# below, as well as the comments and examples in games.d/base.yml for more
# information on the format and how it is used. Examples from production
# servers are also available in branches of the repository:
#   https://github.com/crawl/dgamelaunch-config
# (These are further templated using the scripts in that package.)

# to override yml loading, simply set `games` in this file; see the example
# below. To force using both sources, set `use_game_yaml` to True. Games
# defined in this file will precede games defined in games.d. Templates
# defined in this file are always loaded.
# use_game_yaml = True

# Templating game configurations:
# You can define a set of parameters to use as a template across multiple
# game configs. To use a template by name, use the `template` parameter.
# Recursive templating is supported. If a template named `default` is defined,
# it will be used as the template for any game definitions that specify no
# template at all (see example below). If `default` is defined, it can be
# explicitly overridden by using the template name `base` (which cannot be
# redefined).

# Templating string values:
# if the `version` key is set, most game parameters will support templating
# with %v, %V, and %r.
#   %v: the version value as-is
#   %V: the capitalized version value (e.g. "trunk" -> "Trunk")
#   %r: the 'raw' version, which strips off an initial "0.". So for example,
#       "0.29" would give just "29". If this substring does not occur, %r
#       is equivalent to %v. (Implementation note: characters before a "0."
#       will be dropped as well. Not relevant for normal crawl versions of
#       course.)
#
# Most fields (though not all, in particular, not any game-general ones and
# not socket_path) also support templating with %n, which gives the player's
# username.
#
# The bare minimum necessary fields are what it takes to get the crawl binary
# running. Games are called with the following argv; some of these fields can
# be inferred when not set.
# [
#   $crawl_binary,
#   *$pre_options,
#   "-name", "%n",
#   "-rc", "$rcfile_path/%n.rc",
#   "-macro", "$macro_path/%n.macro",
#   "-morgue", "$morgue_path,"
#   *$options,
#   "-dir", "$dir_path"
#   "-webtiles-socket", "$socket_path/%n:$timestamp.sock",
#   "-await-connection"
# ]
#
# The example below illustrates basic uses of both of these.

# Example of defining `games` directly via a default template and a dictionary,
# and sets all the required paths:
# templates = dict(
#     default = dict(
#         crawl_binary = "./crawl", # relative paths are relative to CWD from running server.py
#         rcfile_path = "./rcs/",
#         macro_path = "./rcs/",
#         morgue_path = "./rcs/%n",
#         inprogress_path = "./rcs/running",
#         ttyrec_path = "./rcs/ttyrecs/%n",
#         socket_path = "./rcs",
#         client_path = "./webserver/game_data/",
#         morgue_url = "http://crawl.akrasiac.org/rawdata/%n/",
#         # the `dir_path` field uses local crawl defaults when unset. This
#         # should almost always be set on a production server, but not for
#         # development.
#         # dir_path = ".",
#         # cwd = ".", # override the CWD inherited when starting the server
#         # morgue_url = None, # public-facing URL, if set
#         show_save_info = True,
#         allowed_with_hold = True,
#         # milestone_path = "./rcs/milestones",
#         # env = {"LANG": "en_US.UTF8"},
#         ),
#     )

# # `games` needs to be an OrderedDict pre python 3.6, so this example uses that
# # for compatibility. Needs the default template above.
# import collections
# games = collections.OrderedDict([
#     ("dcss-web-trunk",   dict(version = "trunk", name="Play %v",)),
#     ("seeded-web-trunk", dict(version = "trunk", name="Seeded", options=["-seed"])),
#     ("tut-web-trunk",    dict(version = "trunk", name="Tutorial", options=["-tutorial"])),
#     ("sprint-web-trunk", dict(version = "trunk", name="Sprint %v", options=["-sprint"])),
# ])
#
# Non-templated example: it's possible just to set every field directly.
# games = collections.OrderedDict([
#     ("dcss-web-trunk", dict(
#         version = "trunk",
#         name = "Play %v",
#         crawl_binary = "./crawl",
#         rcfile_path = "./rcs/",
#         macro_path = "./rcs/",
#         morgue_path = "./rcs/%n",
#         inprogress_path = "./rcs/running",
#         ttyrec_path = "./rcs/ttyrecs/%n",
#         socket_path = "./rcs",
#         client_path = "./webserver/game_data/",
#         morgue_url = None,
#         show_save_info = True,
#         allowed_with_hold = True,
#         # milestone_path = "./rcs/milestones",
#         )),
# ])

dgl_status_file = "./rcs/status"

# Extra paths to tail for milestone updates. This is a legacy setting, you
# should use `milestone_path` or `dir_path` for each game in the games dict.
# (This setting can be a string or list of strings.)
# milestone_file = ["./milestones"]

# status_file_update_rate = 30
# lobby_update_rate = 2

# recording_term_size = (80, 24)

# max_connections = 100

# Script to initialize a user, e.g. make sure the paths
# and the rc file exist. This is not done by the server
# at the moment. This value
init_player_program = "./util/webtiles-init-player.sh"

# how often to check for an active connection while playing; this has an effect
# on how often a connection is checked for basic life, as well as the following
# two idle timer settings.
# connection_timeout = 10 * 60

# the maximum allowed idle time while playing. This timer is checked in
# intervals determined by `connection_timout`, so values will essentially be
# rounded up to that setting
# max_idle_time = 5 * 60 * 60

# the maximum allowed idle time in the lobby. Negative values will disable the
# lobby idle timer. This timer is rounded up to the value of the
# `connection_timeout` setting, similar to `max_idle_time`.
# max_lobby_idle_time = 3 * 60 * 60

# use_gzip = True

# Seconds until stale HTTP connections are closed
# This corresponds to the tornado parameter `idle_connection_timeout`, which
# will automatically close idle http connections that do not respond after a
# period of time. Setting this to `None` gives the tornado default (1 hour).
# This setting usually does not affect websockets connections, which use the
# above webtiles-internal timeouts.
# http_connection_timeout = None

# Set this to true if you are behind a reverse proxy
# Your proxy must set header X-Real-IP
#
# Enabling this option when webtiles is NOT protected behind a reverse proxy
# introduces a security risk. An attacker could inject a false address into the
# X-Real-IP header. Do not enable this option if the webtiles server is
# directly exposed to users.
# http_xheaders = None

# kill_timeout = 10 # Seconds until crawl is killed after HUP is sent

# max_passwd_length = 20

# Username management options.
#
# This options are used on a running webserver only for new account creation!
# It is possible to run a ban query against an existing userdb with them as
# well, see wtutil.py options --check-config-bans and --run-config-bans. For
# banning offensive names, skip to `banned` below. Crawl nicknames are case
# preserving but not case sensitive. To be an acceptable username, a string
# must pass all three of these checks (if defined).
#
# 1. Use `nick_regex` to modify the basic format for valid nicknames.
# a nick is allowed if it matches this regex. It is not recommended to modify
# this unless you really know what you're doing, and this default should not
# ever be modified to something less restrictive for official servers.
# nick_regex = r"^[a-zA-Z0-9]{3,20}$"
#
# 2. `nick_check_fun` if defined, is a function that returns true on valid
# nicknames. You can use this for arbitrary custom nick checks in a server
# config. You will need to do case management manually in this function.
# def nick_check_fun(s):
#     return s.lower() != "plog" and s.lower() != "muggle"
#
# 3. For easily excluding a larger list of usernames from account creation, you
# can use the `banned` option. This can take a simple list of names, or some
# more complex things. Instead of or in addition to putting banned names in
# config, you can put them in files named `banned_players.txt` and
# `banned_players.yml`. The latter must have a single dict entry `banned`
# that is allowed to use the complex options below.
#
# Simple example:
# banned = ['plog', 'muggle', 'muggles', 'ihatemuggles']
#
# Complex example. This will ban many things, for example,
# plain 'muggle', leetspeak variants like 'mugg13s', versions of the phrase
# embedded into other contexts with different letter repeats such as
# 'xx1hatemugggglesxx', etc. It is recommended to be fairly cautious when using
# these extended options, because of
# https://en.wikipedia.org/wiki/Scunthorpe_problem; there are very few English
# slurs where it is a good idea to turn on all three of these options.
# You can mix plain strings with complex entries, as shown here:
#
# banned = ['plog',
#           {'options': {'leet': True, 'repeats': True, 'part': True},
#            'names': ['muggle']}]
# it is allowed to mix in plain strings with dicts in this list,

# Example to demonstrate this same list in yml form:
# banned:
#   - plog
#   - options:
#       leet: True
#       repeats: True
#       part: True
#     names:
#       - muggle
#
# It is recommended to use an external file rather than defining this option
# directly in config. (Among other reasons, this means that server admins are
# only viewing/editing a file likely filled with slurs when they are actually
# intending to...)

# Set to True to allow users to request a password reset email. Some settings
# must be properly configured for this to work:
# allow_password_reset = False
# Set to True to allow dgl admin users to generate password reset tokens in the
# admin panel. Only use if you really trust your admin users!
# admin_password_reset = False

# Set to the primary URL where a player would reach the main lobby
# For example: "http://crawl.akrasiac.org/"
# This is required for for password reset, as it will be the base URL for
# recovery URLs. Use "http://localhost:8080/" for testing.
lobby_url = None

# Proper SMTP settings are required for password reset to function properly.
# if smtp_host is anything other than `localhost`, you may need to adjust the
# timeout settings (see server.py, calls to ioloop.set_blocking_log_threshold).
# TODO: set_blocking_log_threshold is deprecated in tornado 5+...
# Ideally, test out these settings carefully in a non-production setting
# before enabling this, as there's a bunch of ways for this to go wrong and you
# don't want to get your SMTP server blocklisted.
smtp_host = "localhost"
smtp_port = 25
smtp_use_ssl = False
smtp_user = "" # set to None for no auth
smtp_password = ""
smtp_from_addr = "noreply@crawl.example.org" # The address from which automated
                                             # emails will be sent

# crypt() algorithm, e.g. "1" for MD5 or "6" for SHA-512; see crypt(3). If
# false, use traditional DES (but then only the first eight characters of the
# password are significant). If set to "broken", use traditional DES with
# the password itself as the salt; this is necessary for compatibility with
# dgamelaunch, but should be avoided if possible because it leaks the first
# two characters of the password's plaintext.
# crypt_algorithm = "broken"

# The length of the salt string to use. If crypt_algorithm is false, this
# setting is ignored and the salt is two characters.
# crypt_salt_length = 16

# login_token_lifetime = 7 # Days

uid = None  # If this is not None, the server will setuid to that (numeric) id
gid = None  # after binding its sockets.

umask = None # e.g. 0077

chroot = None

pidfile = None
# daemon = False # If true, the server will detach from the session after startup

# Set to a URL with %s where lowercased player name should go in order to
# hyperlink WebTiles spectator names to their player pages.
# For example: "http://crawl.akrasiac.org/scoring/players/%s.html"
# Set to None to disable player page hyperlinks
player_url = None

# set one of these for various moderation modes. Disabled preempts hold. In
# account hold mode, new accounts cannot use chat, cannot spectate, and do
# not appear in the lobby until explicitly approved by an admin. They can still
# play. (Of course, they can still log out and spectate as anon.)
# new_accounts_disabled = True
# new_accounts_hold = True

# customize chat limits. Set to 0 or False to disable chat. Value in characters.
# max_chat_length = 1000

# If set to True, a SIGHUP triggers an attempt to reload the config and game
# data. Some values cannot be reloaded (including this one), and to reset a
# value to its default, you need to explicitly set the value rather than
# comment it out.
# If not explicitly set, this defaults to False.
hup_reloads_config = True

# set to do periodic logging of user load
# load_logging_rate = 10 # seconds, set to 0 to explicitly disable
# slow_callback_alert = 0.25 # seconds, set to None to explicitly disable

# Only for development:
# This is insecure; do not set development_mode = True in production!
# development_mode = False

# Disable caching of static files which are not part of game data.
# no_cache = development_mode
# Automatically log in all users with the username given here.
autologin = None



import datetime
import random
import time

import tornado.ioloop
import tornado.websocket
from tornado.ioloop import IOLoop

from webtiles import config

try:
    from typing import Dict, Tuple
except ImportError:
    pass

login_tokens = {}  # type: Dict[Tuple[str,str],datetime.datetime]
rand = random.SystemRandom()


def purge_login_tokens():
    for token in list(login_tokens):
        if datetime.datetime.now() > login_tokens[token]:
            del login_tokens[token]


def purge_login_tokens_timeout():
    purge_login_tokens()
    IOLoop.current().add_timeout(time.time() + 60 * 60 * 1000,
                                 purge_login_tokens_timeout)


def log_in_as_user(request, username):
    token = rand.getrandbits(128)
    expires = datetime.datetime.now() + datetime.timedelta(config.get('login_token_lifetime'))
    login_tokens[(token, username)] = expires
    cookie = username + "%20" + str(token)
    if not isinstance(request, tornado.websocket.WebSocketHandler):
        request.set_cookie("login", cookie)
    return cookie


def _parse_login_cookie(cookie):
    username, _, token = cookie.partition('%20')
    try:
        token = int(token)
    except ValueError:
        token = None
    return username, token


def check_login_cookie(cookie):
    username, token = _parse_login_cookie(cookie)
    ok = (token, username) in login_tokens
    return username, ok


def forget_login_cookie(cookie):
    username, token = _parse_login_cookie(cookie)
    if (token, username) in login_tokens:
        del login_tokens[(token, username)]



#!/usr/bin/env python3

"""
Generate form-data.h

Works with both Python 2 & 3. If that changes, update how the Makefile calls
this.
"""

from __future__ import print_function

import argparse
import os
import sys
import traceback
import re
import collections
if sys.version_info.major == 2:
    from collections import MutableMapping
else:
    from collections.abc import MutableMapping

from genutil import *

import yaml  # pip install pyyaml

class Field:
    def __init__(self, f, required = False):
        self.f = f
        self.required = required

class Form(MutableMapping):
    """Parser for YAML definition files.

    If any YAML content is invalid, the relevant parser function below should
    raise ValueError.
    """

    def __init__(self, yaml_dict, enum_list):
        self.backing_dict = dict()
        self.from_yaml(yaml_dict, enum_list)

    def __getitem__(self, key):
        return self.backing_dict[key]

    def __setitem__(self, key, val):
        self.backing_dict[key] = val

    def __delitem__(self, key):
        del self.backing_dict[key]

    def __iter__(self):
        return iter(self.backing_dict)

    def __len__(self):
        return len(self.store)

    def from_yaml(self, s, enum_list):
        for key in s:
            if key not in keyfns:
                raise ValueError("unknown key '%s'" % key)
            if key in self:
                raise ValueError("duplicate key '%s'" % key)
            try:
                self[key] = keyfns[key].f(s[key])
            except ValueError as e:
                raise ValueError("failed to parse '%s': %s" % (key, e))

        for key in keyfns:
            if key not in self and keyfns[key].required:
                raise ValueError("missing required key '%s'" % key)

        if self['enum'] not in enum_list:
            raise ValueError("%s does not match any enum" % self['enum'])

        # Derive some names from other names, if none are overrided
        if 'short_name' not in self:
            self['short_name'] = self['enum'].capitalize()
        if 'wiz_name' not in self:
            self['wiz_name'] = self['enum'].lower().replace("_", " ")
        if 'long_name' not in self:
            self['long_name'] = self['short_name'].lower() + "-form"

        for key in defaults:
            if key not in self:
                self[key] = defaults[key]

        self["full_enum"] = "transformation::" + self['enum']
        self["joined_resists"] = ' | '.join(r.format() for r in self["resists"])
        self["ac_scaling"] = self["ac"].format()
        self["ev_scaling"] = self["ev"].format()
        self["body_ac_mult_scaling"] = self["body_ac_mult"].format()
        self["unarmed_scaling"] = self["unarmed"].format()
        self["special_dice"] = self["special_damage"].format()

        if 'TAG_MAJOR_VERSION' in s:
            self['tag_major_version_opener'] = (
                        "\n#if TAG_MAJOR_VERSION == %s" % s['TAG_MAJOR_VERSION'])
            self['tag_major_version_closer'] = "#endif\n"
        else:
            self['tag_major_version_opener'] = ''
            self['tag_major_version_closer'] = ''

def get_fields(d, names):
    fields = []
    for f in names:
        if f not in d:
            raise ValueError("missing field '%s'", f)
        fields.append(d[f])
    if len(d) != len(names):
        raise ValueError("unexpected fields") # XXX add more context
    return fields

def parse_bool(s):
    if not isinstance(s, bool):
        raise ValueError("isn't a bool")
    return 'true' if s else 'false'

def parse_str(s):
    if not isinstance(s, str):
        raise ValueError("isn't a string")
    return s

def parse_str_list(s):
    for a in s:
        if not isinstance(a, str):
            raise ValueError("isn't a string")
    return ', '.join(quote(a) for a in s)

def parse_attack_verbs(s):
    if type(s) is list:
        if len(s) != 1 and len(s) != 4:
            raise ValueError("isn't 1 or 4 strings long")
        return "{ " + parse_str_list(s) + " }"
    if type(s) is str:
        return s.upper() + "_VERBS"
    else:
        raise ValueError("isn't a list or an enum name")

def parse_muts(s):
    if type(s) is not list:
        raise ValueError("isn't a list")
    ret = []
    for a in s:
        ret.append("{" + parse_str_list(a) + "}")
    return ',\n     '.join(ret)

def parse_num(s, min, max):
    if not isinstance(s, int):
        raise ValueError("isn't an int")
    if not min <= s <= max:
        raise ValueError("isn't between %s and %s" % (min, max))
    return s

def parse_flags(s):
    flags = []
    for f in s:
        flags.append("M_" + f.upper())
    return flags

def parse_resists(s):
    resists = []
    for r in s:
        resists.append(parse_resist(r, s[r]))
    return resists

class Resist:
    def __init__(self, min = 0, max = 1):
        self.min = min;
        self.max = max;

ALL_RESISTS = {
    'fire': Resist(-1, 4),
    'cold': Resist(-1, 4),
    'elec': Resist(-1, 4),
    'poison': Resist( -1, 4),
    'neg': Resist(0, 4),
    'corr': Resist(0, 4),
    'miasma': Resist(0, 4),
    'torment': Resist(),
    'petrify': Resist(),
    'damnation': Resist(),
    'steam': Resist(),
}

class ResVal:
    def __init__(self, enum, val):
        self.enum = enum
        self.val = val

    def format(self):
        if self.val == 0 or self.val == 1:
            return self.enum
        return "mrd(%s, %d)" % (self.enum, self.val)

def parse_resist(name, value):
    if not name in ALL_RESISTS:
        raise ValueError("unknown resist '%s', name")

    res = ALL_RESISTS[name]
    try:
        val = parse_num(value, res.min, res.max)
    except ValueError as e:
        raise ValueError("invalid value for resist '%s': %s" % (name, e))
    return ResVal("MR_RES_%s" % name.upper(), val)

class Skill:
    def __init__(self, min, max):
        self.min = min;
        self.max = max;

def parse_skill(s):
    char, colour = get_fields(s, ["min", "max"])
    return Skill(parse_num(s["min"], 0, 27), parse_num(s["max"], 0, 27))

# XXX: Based off equipment_slot. See also transform.cc
ALL_SLOTS = {
    'weapon': 1 << 1,
    'offhand': 1 << 2 | 1 << 12,
    'body': 1 << 3,
    'helmet': 1 << 4,
    'gloves': 1 << 5,
    'boots': 1 << 6,
    'barding': 1 << 7,
    'cloak': 1 << 8,
    'ring': 1 << 9,
    'amulet': 1 << 10,
    'held': 1 << 1 | 1 << 2 | 1 << 12,
    'aux': 1 << 4 | 1 << 5 | 1 << 6 | 1 << 7 | 1 << 8,
    'physical': 1 << 1 | 1 << 2 | 1 << 3 | 1 << 4 | 1 << 5 | 1 << 6 | 1 << 7 | 1 << 8 | 1 << 12,
    'jewellery': 1 << 9 | 1 << 10,
    'all': 1 << 1 | 1 << 2 | 1 << 3 | 1 << 4 | 1 << 5 | 1 << 6 | 1 << 7 | 1 << 8 | 1 << 9 | 1 << 10 | 1 << 12,
}

def parse_slots(s):
    slot = 0
    for f in s:
        if not f in ALL_SLOTS:
            raise ValueError("unknown slot type '%s', f")
        slot |= ALL_SLOTS[f]
    return slot

def parse_enum(s, prefix, vals):
    if s not in vals:
        raise ValueError("%s not in %s" % (s, vals))
    return prefix + s.upper()

SIZES = {'tiny', 'little', 'small', 'medium', 'large', 'giant', 'character'}
def parse_size(s):
    return parse_enum(s, "SIZE_", SIZES)

class FormScaling:
    def __init__(self, base=0, scaling=0, xl_based=False):
        self.base = base;
        self.scaling = scaling;
        self.xl_based = xl_based;

    def format(self):
        if self.base == 0 and self.scaling == 0 and self.xl_based == False:
            return '{}'

        ret = "FormScaling()"
        if self.base != 0:
            ret += ".Base(" + str(self.base) + ")"
        if self.scaling > 0:
            ret += ".Scaling(" + str(self.scaling) + ")"
        if self.xl_based == "true":
            ret += ".XLBased()"
        return ret

def parse_scaling(s):
    return FormScaling(
                base = parse_num(s['base'], -100, 100) if 'base' in s else 0,
                scaling = parse_num(s['scaling'], 0, 100    ) if 'scaling' in s else 0,
                xl_based = parse_bool(s['xl_based'] if 'xl_based' in s else False))

class DiceDef:
    def __init__(self, num=0, size=0, mult=0, div=0):
        self.num = num;
        self.size = size;
        self.mult = mult;
        self.div = div;

    def format(self):
        if self.num == 0 and self.size == 0 and self.mult == 0 and self.div == 0:
            return 'nullptr'
        return "new dicedef_calculator<{},{},{},{}>".format(self.num, self.size, self.mult, self.div)

# Extremely simple/dumb parser that just lets you write dice formulas in a more readable way
def parse_dice(s):
    s = s.replace('d', ',')
    s = s.replace('+', ',')
    s = s.replace('(', ' ')
    s = s.replace(')', ' ')
    s = s.replace('/', ',')
    arg_str = s.split(',')
    args = [0, 0, 0, 1]

    for i in range(0, 4):
        if len(arg_str) > i:
            args[i] = parse_num(int(arg_str[i].strip()), -100, 1000)

    return DiceDef(args[0], args[1], args[2], args[3])

CAPABILITIES = {'default', 'enable', 'forbid'}
def parse_capability(s):
    if not isinstance(s, bool):
        raise ValueError("not a boolean value")
    if s:
        return "FC_ENABLE"
    else:
        return "FC_FORBID"

keyfns = {
    'enum': Field(parse_str, required = True),
    'equivalent_mons': Field(lambda s: "MONS_" + s.upper()),
    'short_name': Field(parse_str),
    'long_name': Field(parse_str),
    'wiz_name': Field(parse_str),

    'description': Field(parse_str),

    'skill': Field(parse_skill),
    'hp_skill_penalty_mult': Field(lambda s: parse_num(s, 1, 10)),
    'talisman': Field(lambda s: "TALISMAN_" + s.upper()),

    'melds': Field(parse_slots),
    'resists': Field(parse_resists),

    'str': Field(lambda s: parse_num(s, -99, 127)),
    'dex': Field(lambda s: parse_num(s, -99, 127)),
    'size': Field(parse_size),
    'hp_mod': Field(lambda s: parse_num(s, 1, 200)),
    'move_speed': Field(lambda s: parse_num(s, 1, 100)),

    'ac': Field(parse_scaling),
    'ev': Field(parse_scaling),
    'body_ac_mult': Field(parse_scaling),
    'can_cast': Field(parse_bool),
    'unarmed': Field(parse_scaling),

    'unarmed_brand': Field(lambda s: "SPWPN_" + s.upper()),
    'unarmed_colour': Field(lambda s: s.upper()),
    'unarmed_name': Field(parse_str),
    'unarmed_verbs': Field(parse_attack_verbs),

    'can_fly': Field(parse_capability),
    'can_swim': Field(parse_capability),
    'offhand_punch': Field(parse_capability),
    'changes_anatomy': Field(parse_bool),
    'changes_substance': Field(parse_bool),
    'holiness': Field(lambda s: "MH_" + s.upper()),

    'is_badform': Field(parse_bool),
    'has_blood': Field(parse_capability),
    'has_hair': Field(parse_capability),
    'has_bones': Field(parse_capability),
    'has_feet': Field(parse_capability),
    'has_ears': Field(parse_capability),

    'shout_verb': Field(parse_str),
    'shout_volume': Field(lambda s: parse_num(s, -100, 100)),
    'hand_name': Field(parse_str),
    'foot_name': Field(parse_str),
    'prayer_action': Field(parse_str),
    'flesh_name': Field(parse_str),

    'special_damage': Field(parse_dice),
    'special_damage_name': Field(parse_str),

    'fakemuts': Field(parse_muts),
    'badmuts': Field(parse_muts),

    'TAG_MAJOR_VERSION': Field(lambda s: parse_num(s, 0, 100)),
}

defaults = {
    'equivalent_mons': 'MONS_PROGRAM_BUG',
    'short_name': "",
    'long_name': "",
    'wiz_name': "",

    'description': "",

    'skill': Skill(0, 0),
    'hp_skill_penalty_mult': 1,
    'talisman': 'NUM_TALISMANS',

    'melds': 0,
    'resists': [ResVal('MR_NO_FLAGS', 0)],

    'str': 0,
    'dex': 0,
    'size': "SIZE_CHARACTER",
    'hp_mod': 100,
    'move_speed': 10,

    'ac': FormScaling(),
    'ev': FormScaling(),
    'body_ac_mult': FormScaling(),
    'can_cast': 'true',
    'unarmed': FormScaling(),

    'unarmed_brand': "SPWPN_NORMAL",
    'unarmed_colour': "LIGHTGREY",
    'unarmed_name': "",
    'unarmed_verbs': "DEFAULT_VERBS",

    'can_fly': "FC_DEFAULT",
    'can_swim': "FC_DEFAULT",
    'offhand_punch': "FC_DEFAULT",
    'changes_anatomy': "false",
    'changes_substance': "false",
    'holiness': "MH_NONE",

    'is_badform': 'false',
    'has_blood': "FC_DEFAULT",
    'has_hair': "FC_DEFAULT",
    'has_bones': "FC_DEFAULT",
    'has_feet': "FC_DEFAULT",
    'has_ears': "FC_DEFAULT",

    'shout_verb': "",
    'shout_volume': 0,
    'hand_name': "",
    'foot_name': "",
    'prayer_action': "",
    'flesh_name': "",

    'special_damage': DiceDef(),
    'special_damage_name': "",

    'fakemuts': "",
    'badmuts': "",
}

def load_template(templatedir, name):
    return open(os.path.join(templatedir, name)).read()

def main():
    parser = argparse.ArgumentParser(description='Generate form-data.h')
    parser.add_argument('datadir', help='dat/forms source dir')
    parser.add_argument('templatedir', help='util/form-gen template source dir')
    parser.add_argument('form_enum', help='transformation.h file path')
    parser.add_argument('form_data', help='form-data.h output file path')
    args = parser.parse_args()

    if not os.path.isdir(args.datadir):
        print('datadir isn\'t a directory')
        sys.exit(1)
    if not os.path.isdir(args.templatedir):
        print('templatedir isn\'t a directory')
        sys.exit(1)

    text = load_template(args.templatedir, 'header.txt')

    tmpl = load_template(args.templatedir, 'body.txt')

    # Determine enum order, so that form data can be output in matching order.
    # This makes a bunch of assumptions about the formating of transformation.h,
    # but those assumptions seem unlikely to change in the near future.
    enum_order = {}
    enum_started = False
    index = 0
    transformation_h = open(args.form_enum).read()
    enum_lines = transformation_h.splitlines()
    for ln in enum_lines:
        trimmed = ln.lstrip()
        if trimmed.startswith("{"):
            enum_started = True
        elif trimmed.startswith("{"):
            enum_started = False
        elif enum_started == False or trimmed.startswith("#") or trimmed.startswith("/"):
            continue
        else:
            enum_name = re.split(',|/s+', trimmed)[0]
            enum_order.update({enum_name : index})
            index += 1

    form_data = []
    for f_name in sorted(os.listdir(args.datadir)):
        if not f_name.endswith('.yaml'):
            continue
        f_path = os.path.join(args.datadir, f_name)
        try:
            form_spec = yaml.safe_load(open(f_path))
        except yaml.YAMLError as e:
            print("Failed to load %s: %s" % (f_name, e))
            sys.exit(1)
        try:
            form_data.append(Form(form_spec, enum_order))
        except (ValueError, KeyError) as e:
            print("Failed to load %s" % f_name)
            traceback.print_exc()
            sys.exit(1)

    # Now sort the data in the proper place
    sorted_form_data = [0] * len(form_data)
    for f in form_data:
        index = enum_order[f['enum']]
        sorted_form_data[index] = f

    for f in sorted_form_data:
        text += tmpl.format(**f)

    text += load_template(args.templatedir, 'footer.txt')

    with open(args.form_data, 'w') as f:
        f.write(text)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""
Generate status-icon-sizes.h, and status-icon-sizes.js

Works with both Python 2 & 3. If that changes, update how the Makefile calls
this.
"""

from __future__ import print_function

import sys

def parse_icon_sizes(file_name, icon_sizes):
    with open(file_name) as input_file:
        line_num = 0
        for line in input_file:
            line_num += 1
            trimmed_line = line.strip()
            if trimmed_line == '' or trimmed_line.startswith('#'):
                continue
            values = trimmed_line.split(':', 1)
            if len(values) < 2:
                print('Error: failed to parse line ', line_num, ' of ',
                      file_name, file=sys.stderr)
                return False

            icon_name = values[0].strip()
            icon_width = values[1].strip()
            if icon_width in icon_sizes:
                icon_sizes[icon_width].append(icon_name)
            else:
                icon_sizes[icon_width] = [icon_name]

    return True

def output_icon_sizes(icon_sizes, input_file_name):
    with open('rltiles/status-icon-sizes.h', 'w') as file:
        file.write('#pragma once\n')
        file.write('\n')
        file.write('// This file is auto generated, edit ' + input_file_name
                   + ' instead.\n')
        file.write('\n')
        file.write('#include "tiledef-icons.h"\n')
        file.write('\n')
        file.write('inline int status_icon_size(tileidx_t icon)\n')
        file.write('{\n')
        file.write('    switch (icon)\n')
        file.write('    {\n')
        for size, names in icon_sizes.items():
            for name in names:
                file.write('    case TILEI_' + name.upper() + ':\n')
            file.write('        return ' + size + ';\n')
        file.write('    default:\n')
        file.write('        return -1;\n')
        file.write('    }\n')
        file.write('}\n')

    with open('rltiles/status-icon-sizes.js', 'w') as file:
        file.write('define(["./tileinfo-icons"], function (icons) {\n')
        file.write('// This file is auto generated, edit ' + input_file_name
                   + ' instead.\n')
        file.write('\n')
        file.write('var exports = {};\n')
        file.write('\n')
        file.write('exports.status_icon_size = function (icon)\n')
        file.write('{\n')
        file.write('    switch (icon)\n')
        file.write('    {\n')
        for size, names in icon_sizes.items():
            for name in names:
                file.write('    case icons.' + name.upper() + ':\n')
            file.write('        return ' + size + ';\n')
        file.write('    default:\n')
        file.write('        return -1;\n')
        file.write('    }\n')
        file.write('}\n')
        file.write('\n')
        file.write('return exports;\n')
        file.write('});\n')

def main():
    if len(sys.argv) < 2:
        print('Error: Missing input file', file=sys.stderr)
        sys.exit(1)

    # List of icons by size
    icon_sizes = {}
    succeeded = parse_icon_sizes(sys.argv[1], icon_sizes)
    if not succeeded:
        sys.exit(1)

    output_icon_sizes(icon_sizes, sys.argv[1])

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Prepare non-C++ files for TAG_MAJOR_VERSION 35 upgrade.

Since most of these files have only basic support for TAG_MAJOR_VERSION, these
edits are done destructively in-place.

You can manually bump the TAG_MAJOR_VERSION by editing tag-version.h.

Bonus: fully remove TAG_MAJOR_VERSION < 35 blocks from C++ files by running:

    util/tag-major-upgrade -t 35
"""

import logging
import os

logging.basicConfig(level=logging.INFO)

OLD_TAG = 34
NEW_TAG = OLD_TAG + 1


def _cleanup_textfile(file_name: str, comment_prefix: str = "#") -> None:
    """Remove old data from a text file with comments.

    We look for lines that are `# start TAG_MAJOR_VERSION == X`. Then behaviour
    until `# end TAG_MAJOR_VERSION` depends on the value of X:
    1. If X == $OLD_TAG, delete all lines
    2. If X == $NEW_TAG, trim `{comment_prefix} ` from each line

    You can customise the comment format, to handle lua files.
    """
    PREFIX_LEN = len(comment_prefix) + 1
    START_MARKER_PREFIX = f"{comment_prefix} start TAG_MAJOR_VERSION == "
    END_MARKER = f"{comment_prefix} end TAG_MAJOR_VERSION\n"
    with open(file_name) as f:
        lines = f.readlines()
    in_tagged_section = None
    with open(file_name, "w") as f:
        for line in lines:
            # If we're in old tag, delete to the end
            if in_tagged_section == OLD_TAG:
                if line == END_MARKER:
                    logging.info("Trimmed old section in %s", file_name)
                    in_tagged_section = None
                continue
            # If we're in new tag, trim the comment prefix
            elif in_tagged_section == NEW_TAG:
                if line == END_MARKER:
                    logging.info("Uncommented new section in %s", file_name)
                    in_tagged_section = None
                    continue
                else:
                    f.write(line[PREFIX_LEN:])
            # If this starts a tag, identify if it's old or new
            elif line.startswith(START_MARKER_PREFIX):
                tag = int(line.strip().rsplit()[-1])
                in_tagged_section = tag
                continue
            # This is a normal line, write it out
            else:
                f.write(line)
        if in_tagged_section is not None:
            raise RuntimeError(f"Didn't find end tag ({END_MARKER!r}) in {file_name}")


def _delete_file(file_name: str) -> None:
    if not os.path.isfile(file_name):
        raise ValueError(f"Asked to delete non-existent file {file_name}")
    logging.info("Deleting %s", file_name)
    os.unlink(file_name)


if __name__ == "__main__":
    SPECIES_BASE = "dat/species"
    for entry in os.scandir(SPECIES_BASE):
        _cleanup_textfile(os.path.join(SPECIES_BASE, entry.name))

    _cleanup_textfile("art-data.txt")
    _cleanup_textfile("util/art-data.pl")

    _cleanup_textfile("rltiles/dc-feat.txt")
    _cleanup_textfile("rltiles/dc-item.txt")
    _cleanup_textfile("rltiles/dc-misc.txt")
    _cleanup_textfile("rltiles/dc-player.txt")

    _cleanup_textfile("dat/des/branches/abyss.des", comment_prefix="--")
    _cleanup_textfile("dat/des/branches/hell.des", comment_prefix="--")
    _cleanup_textfile("dat/des/branches/pan.des", comment_prefix="--")
    # This file has comments in two formats
    _cleanup_textfile("dat/des/branches/tomb.des", comment_prefix="--")
    _cleanup_textfile("dat/des/branches/tomb.des")
    _cleanup_textfile("dat/des/portals/trove.des", comment_prefix="--")
    _cleanup_textfile("dat/des/sprint/zigsprint.des", comment_prefix="--")
    _cleanup_textfile("dat/dlua/v_layouts.lua", comment_prefix="--")
    _delete_file("dat/des/branches/blade.des")
    _delete_file("dat/des/branches/temple_compat.des")
    _delete_file("dat/des/builder/layout_forest.des")
    _delete_file("dat/des/variable/compat.des")

    _delete_file("dat/species/deprecated-draconian-mottled.yaml")

    _delete_file("dat/forms/deprecated-appendage.yaml")
    _delete_file("dat/forms/deprecated-hydra.yaml")
    _delete_file("dat/forms/deprecated-porcupine.yaml")
    _delete_file("dat/forms/deprecated-shadow.yaml")



import codecs
import collections
import datetime
import errno
import logging
import os
import random
import signal
import subprocess
import time
import zlib

import tornado.ioloop
import tornado.template
import tornado.websocket
import tornado.gen
from tornado.escape import json_decode
from tornado.escape import json_encode
from tornado.escape import to_unicode
from tornado.escape import utf8
from tornado.ioloop import IOLoop

from webtiles import auth, checkoutput, config, userdb, util, load_games

try:
    from typing import Dict, Set, Tuple, Any, Union, Optional
except:
    pass

sockets = set() # type: Set[CrawlWebSocket]
current_id = 0
shutting_down = False
rand = random.SystemRandom()

def shutdown():
    global shutting_down
    shutting_down = True
    for socket in list(sockets):
        socket.shutdown()

def update_global_status():
    IOLoop.current().add_callback(write_dgl_status_file)

# lobbies that need updating
game_lobby_cache = set() # type: Set[CrawlProcessHandlerBase]

def list_of_names(l):
    result = []
    try:
        c = collections.Counter(l)
        for n in sorted(c.keys()):
            result.append(n)
            if c[n] > 1:
                result[-1] += " (x%d)" % c[n]
    except:
        pass # backwards compat, py2 doesn't have collections.Counter
    return ", ".join(result)

def describe_sockets(names=False):
    slist = list(sockets)
    lobby = [s for s in slist if s.is_in_lobby()]
    lobby_count = len(lobby)
    # anon connections are either in the lobby or spectating
    anon_lobby = len([s for s in lobby if not s.username])
    anon_specs = len([s for s in slist if not s.username]) - anon_lobby
    playing = [s for s in slist if s.is_running()]
    player_count = len(playing)
    idle_count = len([s for s in playing if s.process.is_idle()])
    spec_count = len([s for s in slist if s.watched_game])

    summary = "%d connections: %d playing (%d idle), %d watching (%d anon), %d in lobby (%d anon)" % (
        len(slist), player_count, idle_count, spec_count, anon_specs, lobby_count, anon_lobby)

    if names:
        # this is all a bit brute-force
        if playing:
            summary += "; Playing: %s" % list_of_names([s.username for s in playing])
        watchers = list_of_names([s.username for s in slist if s.watched_game and s.username])
        if watchers:
            summary += "; Watchers: %s" % watchers
        lobby_names = list_of_names([s.username for s in lobby if s.username and not s.account_restricted()])
        if lobby_names:
            summary += "; Lobby: %s" % lobby_names
        restricted = list_of_names([s.username for s in playing if s.username and s.account_restricted()])
        if restricted:
            summary += "; Account restricted: %s" % restricted
        restricted_lobby = list_of_names([s.username for s in lobby if s.username and s.account_restricted()])
        if restricted_lobby:
            summary += "; Account restricted (lobby): %s" % restricted_lobby
    return summary


def do_lobby_updates():
    global game_lobby_cache
    lobby_sockets = [s for s in list(sockets) if s.is_in_lobby()]
    games_to_update = list(game_lobby_cache)
    game_lobby_cache.clear()
    for game in games_to_update:
        if not game.process:
            # handled immediately in `remove_in_lobbys`, ignore here
            continue
        admin_only = game.account_restricted()

        game_entry = game.lobby_entry()
        # Queue up the collected lobby changes in each socket. This loop is
        # synchronous.
        for socket in lobby_sockets:
            if admin_only and not socket.is_admin():
                continue
            socket.queue_message("lobby_entry", **game_entry)

    # ...and finally, flush all the updates. This loop may be asynchronous.
    for socket in lobby_sockets:
        socket.flush_messages()

def do_periodic_lobby_updates():
    # TODO: in the Future, refactor as a simple async loop
    do_lobby_updates()
    rate = config.get('lobby_update_rate')
    IOLoop.current().add_timeout(time.time() + rate, do_periodic_lobby_updates)

load_logging_enabled = False
def do_load_logging(start = False):
    global load_logging_enabled
    if start and load_logging_enabled:
        # don't start multiple timeouts
        return
    rate = config.get('load_logging_rate')
    if rate:
        load_logging_enabled = True
        if not start:
            # don't log on startup call
            logging.info(describe_sockets())
        IOLoop.current().add_timeout(time.time() + rate, do_load_logging)
    else:
        # config disables load logging, no timeout
        load_logging_enabled = False

def update_all_lobbys(game):
    global game_lobby_cache
    # mark the game for display in the lobby
    game_lobby_cache.add(game)

def remove_in_lobbys(game):
    global game_lobby_cache
    if game in game_lobby_cache:
        game_lobby_cache.remove(game)
    for socket in list(sockets):
        if socket.is_in_lobby():
            socket.send_message("lobby_remove", id=game.id,
                                reason=game.exit_reason,
                                message=game.exit_message,
                                dump=game.exit_dump_url)

def global_announce(text):
    for socket in list(sockets):
        socket.send_announcement(text)

_dgl_dir_check = False


@tornado.gen.coroutine
def write_dgl_status_file():
    process_info = ["%s#%s#%s#0x0#%s#%s#" %
                            (socket.username, socket.game_id,
                             (socket.process.human_readable_where()),
                             str(socket.process.idle_time()),
                             str(socket.process.watcher_count()))
                        for socket in list(sockets)
                        if socket.username and socket.show_in_lobby()]
    try:
        status_target = config.get('dgl_status_file')
        global _dgl_dir_check
        if not _dgl_dir_check:
            status_dir = os.path.dirname(status_target)
            # generally created by other things sooner or later, but if we don't do
            # this preemptively here, there's lot of warnings until that happens.
            if not os.path.exists(status_dir):
                os.makedirs(status_dir)
                logging.warning("Creating dgl status file location '%s'", status_dir)
            _dgl_dir_check = True
        yield util.open_and_write(status_target, "\n".join(process_info))
    except (OSError, IOError) as e:
        logging.warning("Could not write dgl status file: %s", e)


@tornado.gen.coroutine
def status_file_timeout():
    yield write_dgl_status_file()
    IOLoop.current().add_timeout(time.time() + config.get('status_file_update_rate'),
                                 status_file_timeout)


def find_user_sockets(username):
    for socket in list(sockets):
        if socket.username and socket.username.lower() == username.lower():
            yield socket


def find_running_game(charname, start):
    from webtiles.process_handler import processes
    for process in list(processes.values()):
        if (process.where.get("name") == charname and
            process.where.get("start") == start):
            return process
    return None


def _milestone_files():
    # First we collect all milestone files explicitly specified in the config.
    files = set()

    top_level_milestones = config.get('milestone_file')
    if top_level_milestones is not None:
        if not isinstance(top_level_milestones, list):
            top_level_milestones = [top_level_milestones]
        files.update(top_level_milestones)

    for game in config.games:
        milestone_file = config.game_param(game, 'milestone_file')
        if not milestone_file and 'dir_path' in config.games[game]:
            # milestone appears in this dir by default
            milestone_file = os.path.join(
                        config.game_param(game, 'dir_path'), 'milestones')
        if milestone_file:
            files.add(milestone_file)

    # Then, make sure for every milestone we have the -seeded and non-seeded
    # variant.
    new_files = set()
    for f in files:
        if f.endswith('-seeded'):
            new_files.add(f[:-7])
        else:
            new_files.add(f + '-seeded')
    files.update(new_files)

    # Finally, drop any files that don't exist
    files = [f for f in files if os.path.isfile(f)]

    return files

milestone_file_tailers = []
def start_reading_milestones():
    milestone_files = _milestone_files()
    for f in milestone_files:
        milestone_file_tailers.append(util.FileTailer(f, handle_new_milestone))

def handle_new_milestone(line):
    data = util.parse_where_data(line)
    if "name" not in data:
        return
    game = find_running_game(data.get("name"), data.get("start"))
    if game and not game.receiving_direct_milestones:
        game.set_where_info(data)

# decorators for admin calls

# for CrawlWebSocket functions that implement some admin-only user-triggered
# action. These will log if someone tries to call them without permissions.
# Usually for functions that are exposed via a message handler.
def admin_required(f):
    def wrapper(self, *args, **kwargs):
        if not self.is_admin():
            logging.error("Non-admin user '%s' attempted admin function '%s'" %
                (self.username and self.username or "[Anon]", f.__name__))
            return
        return f(self, *args, **kwargs)
    return wrapper

# for CrawlWebSocket functions that should do nothing for a non-admin user, but
# aren't user-triggered. (n.b. it is in principle possible to write a fancier
# decorator that subsumes both of these cases, but it is simpler not to.)
def admin_only(f):
    def wrapper(self, *args, **kwargs):
        if not self.is_admin():
            return
        return f(self, *args, **kwargs)
    return wrapper


# XX add defaults when we can be sure the python version supports it
MessageBundle = collections.namedtuple('MessageBundle', ["binmsg", "compressed"])
MessageBundle.__bool__ = lambda self: bool(self.binmsg)

class CrawlWebSocket(tornado.websocket.WebSocketHandler):
    def __init__(self, app, req, **kwargs):
        tornado.websocket.WebSocketHandler.__init__(self, app, req, **kwargs)
        self.username = None
        self.user_id = None
        self.user_email = None
        self.timeout = None
        self.lobby_timeout = None
        self.watched_game = None
        self.process = None
        self.game_id = None
        self.received_pong = None
        self.save_info = dict()

        tornado.ioloop.IOLoop.current()

        global current_id
        self.id = current_id
        current_id += 1

        self.deflate = True
        self._compressobj = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                             zlib.DEFLATED,
                                             -zlib.MAX_WBITS)
        self.total_message_bytes = 0
        self.compressed_bytes_sent = 0
        self.uncompressed_bytes_sent = 0
        self.message_queue = []  # type: List[str]
        self.failed_messages = 0 # messages to webtiles
        self.failed_on_messages = 0 # messages from webtiles

        self.subprotocol = None

        self.chat_hidden = False

        self.logger = logging.LoggerAdapter(logging.getLogger(), {})
        self.logger.process = self._process_log_msg

        self.message_handlers = {
            "login": self.login,
            "token_login": self.token_login,
            "set_login_cookie": self.set_login_cookie,
            "forget_login_cookie": self.forget_login_cookie,
            "play": self.start_crawl,
            "pong": self.pong,
            "watch": self.watch,
            "chat_msg": self.post_chat_message,
            "register": self.register,
            "start_change_email": self.start_change_email,
            "change_email": self.change_email,
            "start_change_password": self.start_change_password,
            "change_password": self.change_password,
            "forgot_password": self.forgot_password,
            "reset_password": self.reset_password,
            "go_lobby": self.go_lobby,
            "go_admin": self.go_admin,
            "get_rc": self.get_rc,
            "set_rc": self.set_rc,
            "admin_announce": self.admin_announce,
            "admin_pw_reset": self.admin_pw_reset,
            "admin_pw_reset_clear": self.admin_pw_reset_clear
            }

    @admin_required
    def admin_announce(self, text):
        self.reset_lobby_timeout()
        global_announce(text)
        self.logger.info("User '%s' sent serverwide announcement: %s", self.username, text)
        self.send_message("admin_log", text="Announcement made ('" + text + "')")

    @admin_only
    def send_socket_stats(self):
        import webtiles.server
        self.reset_lobby_timeout()
        self.send_message("admin_log", text=webtiles.server.version())
        self.send_message("admin_log", text=describe_sockets(True))

    @admin_required
    def admin_pw_reset(self, username):
        self.reset_lobby_timeout()
        user_info = userdb.get_user_info(username)
        if not user_info:
            self.send_message("admin_pw_reset_done", error="Invalid user")
            return
        ok, msg = userdb.generate_forgot_password(user_info.username)
        if not ok:
            self.send_message("admin_pw_reset_done", error=msg)
        else:
            self.logger.info("Admin user '%s' set a password token on account '%s'",
                self.username, user_info.username)
            self.send_message("admin_pw_reset_done", email_body=msg,
                username=user_info.username, email=user_info.email)

    @admin_required
    def admin_pw_reset_clear(self, username):
        self.reset_lobby_timeout()
        ok, err = userdb.clear_password_token(username)
        if ok:
            self.logger.info("Admin user '%s' cleared the reset token on account '%s'", self.username, username)
        else:
            self.logger.info("Error trying to clear reset token for '%s': %s", username, err)

    client_closed = property(lambda self: (not self.ws_connection) or self.ws_connection.client_terminated)

    def _process_log_msg(self, msg, kwargs):
        return "#%-5s %s" % (self.id, msg), kwargs

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        return self.id == other.id

    def allow_draft76(self):
        return True

    def select_subprotocol(self, subprotocols):
        if "no-compression" in subprotocols:
            self.deflate = False
            self.subprotocol = "no-compression"
            return "no-compression"
        return None

    def open(self):
        compression = "on"
        if isinstance(self.ws_connection, getattr(tornado.websocket, "WebSocketProtocol76", ())):
            # Old Websocket versions don't support binary messages
            self.deflate = False
            compression = "off, old websockets"
        elif self.subprotocol == "no-compression":
            compression = "off, client request"
        if hasattr(self, "get_extensions"):
            if any(s.endswith("deflate-frame") for s in self.get_extensions()):
                self.deflate = False
                compression = "deflate-frame extension"

        self.logger.info("Socket opened from ip %s (fd%s, compression: %s).",
                         self.request.remote_ip,
                         self.ws_connection.stream.socket.fileno(),
                         compression)
        sockets.add(self)

        self.reset_idle_timeouts()

        if config.get('max_connections') < len(sockets):
            self.append_message("connection_closed('The maximum number of "
                              + "connections has been reached, sorry :(');")
            self.close()
        elif shutting_down:
            self.close()
        else:
            if config.get('dgl_mode'):
                if config.get('autologin'):
                    self.do_login(config.get('autologin'))
                self.send_lobby()
            else:
                self.start_crawl(None)

    def check_origin(self, origin):
        return True

    def idle_time(self):
        return self.process.idle_time()

    def is_running(self, partial=False):
        # during setup, and for a short interval after close, self.process
        # is set, but self.process.process may be unset or indeterminate.
        # after close, `try_cleanup` provides a safe point for this check.
        return self.process is not None and (partial or self.process.process)

    def is_in_lobby(self):
        # caveat: this will return True while a player is logging in directly
        # via a play link...
        return not self.is_running(partial=True) and self.watched_game is None

    def send_lobby_data(self):
        self.queue_message("lobby_clear")
        from webtiles.process_handler import processes
        for process in list(processes.values()):
            if process.account_restricted() and not self.is_admin():
                continue
            self.queue_message("lobby_entry", **process.lobby_entry())
        self.send_message("lobby_complete")

    def send_lobby(self):
        self.send_lobby_data()
        self.send_lobby_html()
        self.send_socket_stats() # admins only

    def account_restricted(self):
        return not self.username or userdb.dgl_account_hold(self.user_flags)

    def show_in_lobby(self):
        return (self.is_running()
            and not self.account_restricted())

    def send_announcement(self, text):
        # TODO: something in lobby?
        if not self.is_in_lobby():
            # show in chat window
            self.send_message("server_announcement", text=text)
            # show in player message window
            if self.is_running():
                self.process.handle_announcement(text)

    def invalidate_saveslot_cache(self, slot):
        # TODO: the following will get false positives. However, to do any
        # better would need some non-trivial refactoring of how crawl handles
        # save slots (which is a bit insane). A heuristic might be to check the
        # binary, but in practice this doesn't help as most servers launch
        # crawl via a wrapper script, not via a direct call.
        if self.save_info.get(slot, None) is not None:
            cache_check = "" if self.save_info[slot] == "" else "[slot full]"
            for g in self.save_info:
                if self.save_info[g] == cache_check:
                    self.save_info[g] = None
        self.save_info[slot] = None

    def update_save_info(self, binary_key, data):
        try:
            data = json_decode(data)
        except Exception:
            # JSON error. It can also happen under some dgamelaunch-config
            # setups if escape codes are incorrectly inserted into
            # the output for some calls. See:
            # https://github.com/crawl/dgamelaunch-config/commit/6ad788ceb5614b3c83d65b61bf26a122e592b98d
            # should this case warn? It probably means a misconfiguration.
            for g in config.binaries_to_games[binary_key]:
                self.save_info[g] = ""
                return

        # data is now a dict of game mode to game descriptions. If game modes
        # share a save slot, then they will map to the same game.
        for g in config.binaries_to_games[binary_key]:
            if (not config.game_modes[g] in data
                        or not config.games[g].get("show_save_info", False)):
                self.save_info[g] = ""
                continue
            save_dict = data[config.game_modes[g]]
            if not save_dict["loadable"]:
                # the save in this slot is in use.
                self.save_info[g] = "[playing]" # TODO: something better??
            elif config.game_modes[g] == save_dict["game_type"]:
                # save in the slot matches the game type we are
                # checking.
                self.save_info[g] = "[%s]" % save_dict["short_desc"]
            else:
                # There is a save, but it has a different game type.
                # This happens if multiple game types share a slot.
                self.save_info[g] = "[slot full]"

    # collect save info for the player from all binaries that support save
    # info json. Cached on self.save_info. This is asynchronously done using
    # a somewhat involved callback chain; each check requires a subprocess
    # call on which we don't want to block.
    # this code would be much simpler if refactored using async
    def collect_save_info(self, final_callback):
        if not self.username:
            return

        # Tally up the calls we need in order to check relevant save info.
        # For each distinct binary (going by binary keys, cached in
        # `binaries_to_games`), if that binary has modes that should show and
        # update save info, add a call to the list.
        check_calls = []
        for b in config.binaries_to_games:
            games = config.binaries_to_games[b]
            if not games:
                continue
            if not config.game_modes[games[0]]:
                continue # no game mode info for this game
            if not any([config.games[g].get("show_save_info", False) for g in games]):
                # no games modes for this binary have enabled save info
                for g in games:
                    self.save_info[g] = ""
                continue
            if not any([self.save_info.get(g, None) is None for g in games]):
                # no invalid caches for this binary
                continue
            # doesn't matter which element of games we use
            call = config.games[games[0]].get_call_base()
            call += ["-save-json", self.username]
            check_calls.append([b, call])

        # next we need to turn the list of calls into a sequence of non-blocking
        # callbacks

        def build_callback(b, call, next_callback):
            # build a callback that:
            # a. calls check_output with the call info (nonblocking subprocess)
            # b. passes the output to update_save_info, and
            # c. calls next_callback
            def do_update(data, returncode):
                if returncode != 0:
                    # Binary doesn't support save info, print a warning so that
                    # the admin can see there is a misconfiguration
                    # XX it should be possible to check for this case on startup
                    logging.warning("Save info check for '%s' failed" % b)
                    data = None # force error case in update_save_info call

                self.update_save_info(b, data)
                next_callback()

            # the check_output call is non-blocking, e.g. will yield to other
            # IOLoop stuff:
            return lambda: checkoutput.check_output(call, do_update)

        # sequence the calls into a chain of callbacks that end with
        # final_callback
        callback = final_callback
        for l in check_calls:
            callback = build_callback(l[0], l[1], callback)

        # finally, kick things off
        callback()

    def send_lobby_html(self):
        # Rerender Banner
        # TODO: don't really need to do this every time the lobby is loaded?
        if self.ui is None:
            # socket has been finish()ed, and the render_string call will raise.
            # this probably only happens in async race conditions, so don't
            # do anything further (and hope that the close happens properly
            # elsewhere)
            return
        banner_html = to_unicode(self.render_string("banner.html",
                                                    username = self.username))
        self.queue_message("html", id = "banner", content = banner_html)

        if not self.username:
            return
        def disable_check(s):
            return s == "[slot full]"
        def send_game_links():
            global sockets
            if not self in sockets:
                return # socket closed by the time this info was collected
            self.queue_message("html", id = "banner", content = banner_html)
            # TODO: dynamically send this info as it comes in, rather than
            # rendering it all at the end?
            try:
                # post py3.6, this can all be done with a dictionary
                # comprehension, but before that we need to manually keep
                # the order
                games = collections.OrderedDict()
                for g in config.games:
                    if self.game_id_allowed(g):
                        games[g] = config.games[g].templated_dict(self.username)
                play_html = to_unicode(self.render_string("game_links.html",
                                                  games = games,
                                                  save_info = self.save_info,
                                                  disabled = disable_check))
                self.send_message("set_game_links", content = play_html)
            except:
                self.logger.warning("Error on send_game_links callback",
                                                                exc_info=True)
        # if no game links at all have been sent, immediately render the
        # empty version. This is so that if the server takes a while on
        # initial connect, the player sees something immediately.
        if len(self.save_info) == 0:
            for g in config.games:
                self.save_info[g] = None
            send_game_links()
        self.collect_save_info(send_game_links)

    def clear_timeouts(self, lobby=None):
        if not lobby and self.timeout:
            IOLoop.current().remove_timeout(self.timeout)
            self.timeout = None
        if lobby != False and self.lobby_timeout:
            IOLoop.current().remove_timeout(self.lobby_timeout)
            self.lobby_timeout = None

    def reset_lobby_timeout(self):
        self.clear_timeouts(lobby=True)
        # warning: there is a somewhat painful interval at game start while
        # self.process is partly set up, but before self.process.process is
        # still unset. Don't call this function during that time...
        time_limit = config.get('max_lobby_idle_time')
        if not self.is_in_lobby() or self.is_admin() or time_limit <= 0:
            # this can be triggered on callbacks; if the player is now playing,
            # leave things with this timeout no longer set at all
            return

        # don't check more often than the overall connection timeout setting
        time_limit = max(time_limit, config.get('connection_timeout'))
        # lobby doesn't track idle time, so we just use a timeout directly;
        # this gets reset on actions
        self.lobby_timeout = IOLoop.current().add_timeout(
                        time.time() + time_limit,
                        self.check_lobby_connection)

    def check_lobby_connection(self):
        if self.is_running(): # sanity check, but this shouldn't be possible
            return
        # unconditionally close if we reach the lobby idle time
        # XX: possibly print a more informative message on the client side?
        self.logger.info("Lobby connection timed out.")
        self.close()

    def reset_timeout(self):
        self.clear_timeouts(lobby=False)

        self.received_pong = False
        self.send_message("ping")
        self.timeout = IOLoop.current().add_timeout(
                                        time.time() + config.get('connection_timeout'),
                                        self.check_connection)

    def reset_idle_timeouts(self):
        self.reset_lobby_timeout()
        self.reset_timeout()

    def check_connection(self):
        self.timeout = None

        if not self.received_pong:
            self.logger.info("Connection timed out.")
            self.close()
        elif self.is_running():
            if self.process.idle_time() > config.get('max_idle_time'):
                self.logger.info("Stopping crawl after idle time limit.")
                self.process.stop()
        elif self.is_in_lobby():
            if not self.lobby_timeout:
                self.reset_lobby_timeout()
        # XX none of these covers watchers

        if not self.client_closed:
            self.reset_timeout()

    def update_db_info(self):
        """update self's user info from the user database, including updating
        any flag changes that may have happened in the db."""
        if not self.username:
            return True # caller needs to check for anon if necessary
        # won't detect a change in hold state on first login...
        old_restriction = self.user_flags is not None and self.account_restricted()
        u = userdb.get_user_info(self.username)
        self.username = u.username # canonicalize
        self.user_id = u.id
        self.user_email = u.email
        self.user_flags = u.flags
        self.logger.extra["username"] = self.username
        if userdb.dgl_is_banned(self.user_flags):
            return False
        if old_restriction and not self.account_restricted():
            self.logger.info("[Account] Hold cleared for user %s (IP: %s)",
                                        self.username, self.request.remote_ip)

        return True

    def game_id_allowed(self, game_id):
        if game_id not in config.games:
            return False
        if not self.account_restricted():
            return True
        game = config.games[game_id]
        # for now, if this isn't set, default to allow. However, if the
        # version doesn't support `-no-player-bones` it will still fail to
        # start.
        return "allowed_with_hold" not in game or game["allowed_with_hold"]

    def start_crawl(self, game_id):
        if config.get('dgl_mode') and game_id not in config.games:
            self.go_lobby()
            return

        if config.get('dgl_mode'):
            if self.username == None:
                if self.watched_game:
                    self.stop_watching()
                self.send_message("login_required",
                    game=config.games[game_id].templated("name", username=self.username))
                return

        if self.process:
            # ignore multiple requests for the same game, can happen when
            # logging in with cookies
            if self.game_id != game_id:
                self.go_lobby()
            return

        self.game_id = game_id

        # invalidate cached save info for lobby
        # TODO: invalidate for other sockets of the same player?
        self.invalidate_saveslot_cache(game_id)

        # update flags in case an account hold has been released or added, or
        # the player has been banned.
        if config.get('dgl_mode') and (not self.update_db_info() # ban check happens here
                or not self.game_id_allowed(game_id)):
            self.go_lobby()
            return

        from webtiles import process_handler

        if config.get('dgl_mode'):
            self.process = process_handler.CrawlProcessHandler(
                            config.games[game_id], self.username, self.logger)
        else:
            self.process = process_handler.DGLLessCrawlProcessHandler(self.logger)

        # now that `process` is set, clear the lobby timeout and reset the
        # play timeout
        self.reset_idle_timeouts()
        # self.clear_timeouts(lobby=True)
        self.process.end_callback = self._on_crawl_end
        self.process.add_watcher(self)
        try:
            self.process.start()
        except Exception:
            self.logger.warning("Exception starting process!", exc_info=True)
            self.process = None
            self.go_lobby()
        else:
            if self.process is None: # Can happen if the process creation fails
                # lobby should be handled in exit callback
                return

            self.send_message("game_started")
            self.restore_blocklist()
            self.process.handle_notification(self.process.username,
                            "'/help' to see available chat commands")
            if not config.get('max_chat_length'):
                self.process.handle_notification(self.process.username,
                            "Regular chat is disabled on this server")


            if config.get('dgl_mode'):
                if self.process.where == {}:
                    # If location info was found, the lobbys were already
                    # updated by set_where_data
                    update_all_lobbys(self.process)
                update_global_status()

    def _on_crawl_end(self):
        if self.process:
            if config.get('dgl_mode'):
                remove_in_lobbys(self.process)

            reason = self.process.exit_reason
            message = self.process.exit_message
            dump_url = self.process.exit_dump_url
        else:
            # XX under what circumstance is this reachable?
            reason = None

        self.process = None

        if not self.client_closed:
            if shutting_down:
                self.close()
            else:
                # Go back to lobby
                if reason:
                    self.send_message("game_ended", reason = reason,
                                      message = message, dump = dump_url)

                self.invalidate_saveslot_cache(self.game_id)

                if config.get('dgl_mode'):
                    if not self.watched_game:
                        self.send_message("go_lobby")
                        self.send_lobby()
                else:
                    self.start_crawl(None)

        if config.get('dgl_mode'):
            # queues a callback to update the status list
            update_global_status()

        self.try_cleanup()

    def init_user(self, callback):
        # this would be more cleanly implemented with wait_for_exit, but I
        # can't get code for that to work in a way that supports all currently
        # in-use versions. TODO: clean up once old Tornado versions are out of
        # the picture.
        with open("/dev/null", "w") as f:
            if tornado.version_info[0] < 3:
                # before tornado 3, an async approach would have to be done
                # differently, and given that we're deprecating tornado 2.4
                # it doesn't seem worth implementing right now. Just stick with
                # the old synchronous approach for backwards compatibility.
                p = subprocess.Popen([config.get('init_player_program'), self.username],
                                         stdout = f, stderr = subprocess.STDOUT)
                callback(p.wait())
            else:
                # TODO: do we need to care about the streams at all here?
                p = tornado.process.Subprocess(
                        [config.get('init_player_program'), self.username],
                        stdout = f, stderr = subprocess.STDOUT)
                p.set_exit_callback(callback)

    def stop_watching(self):
        if self.watched_game:
            self.logger.info("%s stopped watching %s.",
                                    self.username and self.username or "[Anon]",
                                    self.watched_game.username)
            self.watched_game.remove_watcher(self)
            self.watched_game = None

    def shutdown(self):
        if not self.client_closed:
            self.logger.info("Shutting down user %s id %d",
                                    self.username and self.username or "[Anon]",
                                    self.id)
            msg = to_unicode(self.render_string("shutdown.html", game=self))
            self.send_message("close", reason = msg)
            self.close()
        if self.is_running():
            self.process.stop()

    def do_login(self, username):
        self.username = username
        self.user_flags = None
        if not self.update_db_info():
            # XX consolidate with other ban check / login fail code somehow.
            # Also checked in userdb.user_passwd_match.
            fail_reason = 'Account is disabled.'
            self.logger.warning("[Account] Failed login for user %s: %s", self.username,
                                    fail_reason)
            self.send_message("login_fail", reason=fail_reason)
            return

        def login_callback(result):
            success = result == 0
            if not success:
                msg = ("Could not initialize your rc and morgue!<br>" +
                       "This probably means there is something wrong " +
                       "with the server configuration.")
                self.send_message("close", reason = msg)
                self.logger.warning("User initialization returned an error for user %s!",
                                    self.username)
                self.username = None
                self.close()
                return

            self.queue_message("login_success", username=username,
                               admin=self.is_admin())
            if self.account_restricted():
                self.queue_message("set_account_hold")
            if self.watched_game:
                self.watched_game.update_watcher_description()
            elif self.is_admin():
                self.send_lobby()
            else:
                self.send_lobby_html()

            self.reset_lobby_timeout()

        self.init_user(login_callback)

    def login(self, username, password):
        success, real_username, fail_reason = userdb.user_passwd_match(username, password)
        if success and real_username:
            self.logger.info("User %s logging in from %s.",
                                        real_username, self.request.remote_ip)
            self.do_login(real_username)
        elif fail_reason:
            self.logger.warning("Failed login for user %s (IP: %s): %s",
                            real_username, self.request.remote_ip, fail_reason)
            self.send_message("login_fail", reason = fail_reason)
        else:
            self.logger.warning("Failed login for user %s (IP: %s).", username,
                            self.request.remote_ip)
            self.send_message("login_fail")

    def token_login(self, cookie):
        username, ok = auth.check_login_cookie(cookie)
        if ok:
            auth.forget_login_cookie(cookie)
            self.logger.info("User %s logging in from %s (via token).",
                username, self.request.remote_ip)
            self.do_login(username)
        else:
            self.logger.warning("Wrong login token for user %s. (IP: %s)",
                username, self.request.remote_ip)
            self.send_message("login_fail")

    def set_login_cookie(self):
        if self.username is None:
            return
        cookie = auth.log_in_as_user(self, self.username)
        self.send_message("login_cookie", cookie = cookie,
                          expires = config.get('login_token_lifetime'))

    def forget_login_cookie(self, cookie):
        auth.forget_login_cookie(cookie)

    def restore_blocklist(self):
        if not self.username:
            return
        receiver = None
        if self.process:
            receiver = self.process
        elif self.watched_game:
            receiver = self.watched_game

        if not receiver:
            return

        db_string = userdb.get_blocklist(self.username)
        if db_string is None:
            db_string = ""
        # list constructor here is for forward compatibility with python 3.
        blocked = list([_f for _f in db_string.strip().split(' ') if _f])
        receiver.restore_blocklist(self.username, blocked)

    def save_blocklist(self, blocked):
        db_string = " ".join(blocked).strip()
        userdb.set_blocklist(self.username, db_string)

    def is_admin(self):
        return self.username is not None and userdb.dgl_is_admin(self.user_flags)

    def pong(self):
        self.received_pong = True

    def rcfile_path(self, game_id):
        if game_id not in config.games:
            return None
        if not self.username:
            return None
        path = config.game_param(game_id, "rcfile_path", username=self.username)
        return os.path.join(path, self.username + ".rc")

    def send_json_options(self, game_id, player_name):
        def do_send(data, returncode):
            if returncode != 0:
                # fail silently for returncode 1 for now, probably just an old
                # version missing the command line option
                if returncode != 1:
                    self.logger.warning("Error while getting JSON options!")
                return
            self.append_message('{"msg":"options","watcher":true,"options":'
                                + data + '}')

        if not self.username: return
        if game_id not in config.games: return

        game = config.games[game_id]
        if "send_json_options" in game and not game["send_json_options"]:
            return

        call = game.get_call_base()

        call += ["-name", player_name,
                 "-rc", self.rcfile_path(game_id)]
        if "options" in game:
            call += game.templated("options", username=self.username)
        call.append("-print-webtiles-options")

        checkoutput.check_output(call, do_send)

    def watch(self, username):
        if self.is_running():
            self.process.stop()

        if not self.update_db_info():
            self.go_lobby()
            return

        if self.username and self.account_restricted():
            self.go_lobby(message="Your account is restricted; spectating unavailable")
            return

        if not self.username and not config.get('allow_anon_spectate'):
            self.send_message("auth_error",
                        reason="Anonymous spectating disabled")
            self.go_lobby()
            return

        from webtiles.process_handler import processes
        procs = [process for process in list(processes.values())
                 if process.username.lower() == username.lower()]
        if len(procs) >= 1:
            process = procs[0]
            # has the spectatee blocked this user?
            if process.is_blocked(self.username) and not self.is_admin():
                self.send_message("auth_error",
                            reason="Spectating this player is restricted.")
                self.go_lobby()
                return
            r = process.get_primary_receiver()
            # the spectatee has a restricted account; you can only get to this
            # by manually typing a url
            if r and r.account_restricted():
                self.go_lobby()
                return

            if self.watched_game:
                if self.watched_game == process:
                    return
                self.stop_watching()
            self.logger.info("%s started watching %s (%s).",
                                self.username and self.username or "[Anon]",
                                process.username, process.id)

            self.watched_game = process
            process.add_watcher(self)
            self.send_message("watching_started", username = process.username)
            if not config.get('max_chat_length'):
                process.handle_notification(self.username,
                                            "Chat is disabled on this server")
        else:
            if self.watched_game:
                self.stop_watching()
            self.go_lobby()

    def post_chat_message(self, text):
        max_length = config.get('max_chat_length')
        if max_length:
            max_length = max(100, int(max_length)) # if it's set, enforce a minimum
            if len(text) >= max_length:
                text = text[:max_length - 5] + "[...]"
        receiver = None
        if self.process:
            receiver = self.process
        elif self.watched_game:
            receiver = self.watched_game

        if receiver:
            if self.account_restricted():
                self.send_message("chat",
                        content='Your account is restricted; chat unavailable')
                return
            if self.username is None:
                self.send_message("chat",
                        content='You need to log in to send messages!')
                return

            if not receiver.handle_chat_command(self, text):
                if not max_length:
                    return # chat disabled!
                receiver.handle_chat_message(self.username, text)

    def register(self, username, password, email):
        error = userdb.register_user(username, password, email)
        if error is None:
            self.logger.info("[Account] Registered user %s (IP: %s, email: %s).",
                            username, self.request.remote_ip, email)
            self.do_login(username)
        else:
            self.logger.info("[Account] Registration attempt failed for user %s (IP: %s): %s",
                             username, self.request.remote_ip, error)
            self.send_message("register_fail", reason = error)

    def start_change_password(self):
        self.reset_lobby_timeout()
        self.send_message("start_change_password")

    def change_password(self, cur_password, new_password):
        self.reset_lobby_timeout()

        if not self.update_db_info():
            self.send_message("change_password_fail", reason="Account is disabled")
            return

        if self.username is None:
            self.send_message("change_password_fail",
                    reason="You need to log in to change your password.")
            return

        if self.account_restricted():
            self.send_message("change_password_fail",
                    reason="Your account is restricted; change password unavailable.")
            return

        if not userdb.user_passwd_match(self.username, cur_password)[0]:
            self.send_message("change_password_fail", reason = "Your password didn't match.")
            self.logger.info("Non-matching current password during password change for %s", self.username)
            return

        error = userdb.change_password(self.user_id, new_password)
        if error is None:
            self.logger.info("User %s changed password.", self.username)
            self.send_message("change_password_done")
        else:
            self.logger.info("Failed to change username for %s: %s", self.username, error)
            self.send_message("change_password_fail", reason = error)


    def start_change_email(self):
        self.reset_lobby_timeout()
        self.send_message("start_change_email", email = self.user_email)

    def change_email(self, email):
        self.reset_lobby_timeout()

        if not self.update_db_info():
            self.send_message("change_email_fail", reason="Account is disabled")
            return

        if self.username is None:
            self.send_message("change_email_fail", reason = "You need to log in to change your email")
            return
        if self.account_restricted():
            self.send_message("change_email_fail",
                    reason="Your account is restricted; change email unavailable.")
            return
        error = userdb.change_email(self.user_id, email)
        if error is None:
            self.update_db_info()
            self.logger.info("User %s changed email to %s.",
                self.username, self.user_email if self.user_email else "null")
            self.send_message("change_email_done", email = self.user_email)
        else:
            self.logger.info("Failed to change username for %s: %s", self.username, error)
            self.send_message("change_email_fail", reason = error)

    def forgot_password(self, email):
        if not config.get("allow_password_reset", False):
            return
        self.reset_lobby_timeout()

        sent, error = userdb.send_forgot_password(email)
        if error is None:
            if sent:
                self.logger.info("Sent password reset email to %s.", email)
            else:
                self.logger.info("User requested a password reset, but email "
                                 "is not registered (%s).", email)
            self.send_message("forgot_password_done")
        else:
            self.logger.info("Failed to send password reset email for %s: %s",
                             email, error)
            self.send_message("forgot_password_fail", reason = error)

    def reset_password(self, token, password):
        self.reset_lobby_timeout()

        username, error = userdb.update_user_password_from_token(token,
                                                                 password)
        if error is None:
            self.logger.info("User %s has completed their password reset.",
                             username)
            self.send_message("reload_url")
        else:
            if username is None:
                self.logger.info("Failed to update password for token %s: %s",
                                 token, error)
            else:
                self.logger.info("Failed to update password for user %s: %s",
                                 username, error)
            self.send_message("reset_password_fail", reason = error)

    def go_lobby(self, message=None):
        if not config.get('dgl_mode'):
            return

        if message:
            # a bit hacky: puts the message above the login box / username
            self.send_message("auth_error", reason=message)

        if self.is_running():
            self.process.stop()

        # note -- lobby timer is not guaranteed to be started here, because
        # process stop is async
        self.reset_idle_timeouts()

        if self.username and userdb.dgl_is_banned(self.user_flags):
            # force a logout. Note that this doesn't check the db at this point
            # in order to reduce i/o a bit.
            fail_reason = 'Account is disabled.'
            self.logger.warning("[Account] Logging out user %s: %s",
                                    self.username, fail_reason)
            self.username = self.user_email = self.user_flags = self.user_id = None
            self.send_message("go_lobby")
            self.send_lobby_html()
            self.send_message("logout", reason=fail_reason)
            return

        if self.watched_game:
            self.stop_watching()
            self.send_message("go_lobby")
            self.send_lobby()
        else:
            self.send_message("go_lobby")

    def go_admin(self):
        self.go_lobby()
        self.send_message("go_admin")

    def get_rc(self, game_id):
        if game_id not in config.games: return

        self.reset_lobby_timeout()

        path = self.rcfile_path(game_id)
        try:
            with util.SlowWarning("Slow IO: read rc '%s'" % path):
                with open(path, 'r') as f:
                    contents = f.read()
        # Handle RC file not existing. IOError for py2, OSError for py3
        except (OSError, IOError):
            contents = ''
        self.send_message("rcfile_contents", game_id = game_id, contents = contents)

    def set_rc(self, game_id, contents):
        rcfile_path = self.rcfile_path(game_id)

        self.reset_lobby_timeout()

        try:
            with util.SlowWarning("Slow IO: write rc '%s'" % rcfile_path):
                with open(rcfile_path, 'wb') as f:
                    # TODO: is binary + encode necessary in py 3?
                    f.write(utf8(contents))
        except Exception:
            self.logger.warning(
                    "Couldn't save rcfile for %s!",
                    self.username, exc_info=True)
            return False
        return True

    def on_message(self, message): # type: (Union[str, bytes]) -> None
        try:
            obj = json_decode(message) # type: Dict[str, Any]
            if obj["msg"] in self.message_handlers:
                handler = self.message_handlers[obj["msg"]]
                del obj["msg"]
                handler(**obj)
            elif self.process:
                self.process.handle_input(message)
            elif (not self.watched_game
                    and obj["msg"] != 'ui_state_sync'
                    and obj["msg"] != 'key'):
                # ui_state_sync can get queued by the js client just before
                # shutdown, and have its sending delayed by enough that the
                # process has stopped. I do not currently think there's a
                # principled way to suppress it with this timing on the js
                # side (because it's basically just a consequence of general
                # ui code), so have an ugly exception here. Otherwise, various
                # game ending conditions log a warning. TODO: fixes?
                self.logger.warning("Didn't know how to handle msg (user %s): %s",
                                    self.username and self.username or "[Anon]",
                                    obj["msg"])
            if self.failed_on_messages > 1:
                self.logger.warning("%d more on_message errors...", self.failed_on_messages)
            self.failed_on_messages = 0
        except OSError as e:
            # maybe should throw a custom exception from the socket call rather
            # than rely on these cases?
            if self.failed_on_messages == 0:
                excerpt = message[:50] if len(message) > 50 else message
                trunc = "..." if len(message) > 50 else ""
                if e.errno == errno.EAGAIN or e.errno == errno.ENOBUFS:
                    # errno is different on mac vs linux, maybe also depending on
                    # python version
                    self.logger.warning(
                        "Socket buffer full; skipping JSON message ('%r%s')!",
                                    excerpt, trunc)
                else:
                    self.logger.warning(
                                    "Error while handling JSON message ('%r%s')!",
                                    excerpt, trunc,
                                    exc_info=True)
            self.failed_on_messages += 1
        except Exception:
            if self.failed_on_messages == 0:
                excerpt = message[:50] if len(message) > 50 else message
                trunc = "..." if len(message) > 50 else ""
                self.logger.warning("Error while handling JSON message ('%r%s')!",
                                    excerpt, trunc,
                                    exc_info=True)
            self.failed_on_messages += 1

    def _encode_for_send(self, msg, deflate):
        try:
            binmsg = utf8(msg)
            if deflate:
                # Compress like in deflate-frame extension:
                # Apply deflate, flush, then remove the 00 00 FF FF
                # at the end
                # note: a compressed stream is stateful, you can't use this
                # compressobj for other sockets
                compressed = self._compressobj.compress(binmsg)
                compressed += self._compressobj.flush(zlib.Z_SYNC_FLUSH)
                compressed = compressed[:-4]
                return MessageBundle(binmsg, compressed)
            else:
                return MessageBundle(binmsg, None)
        except:
            # might happen with weird utf-8 stuff...can this be handled more
            # precisely?
            self.logger.warning("Exception trying to encode message.", exc_info = True)
            return MessageBundle(None, None)


    # send a single message batch, encoding and compressing it if necessary
    def _send_raw_message(self, msg):
        if self.client_closed or not msg:
            return False

        bundle = self._encode_for_send(msg, self.deflate)
        if not bundle:
            self.failed_messages += 1
            return False

        try:
            self.total_message_bytes += len(bundle.binmsg)
            if self.deflate:
                self.compressed_bytes_sent += len(bundle.compressed)
                f = self.write_message(bundle.compressed, binary=True)
            else:
                self.uncompressed_bytes_sent += len(bundle.binmsg)
                f = self.write_message(bundle.binmsg)

            import traceback
            cur_stack = traceback.format_stack()

            # handle any exceptions lingering in the Future
            # TODO: this whole call chain should be converted to use coroutines
            def after_write_callback(f):
                try:
                    f.result()
                except tornado.websocket.StreamClosedError as e:
                    # not supposed to be raised here in current versions of
                    # tornado, but in some older versions it is.
                    if self.failed_messages == 0:
                        self.logger.warning("Connection closed during async write_message")
                    self.failed_messages += 1
                    if self.ws_connection is not None:
                        self.ws_connection._abort()
                except tornado.websocket.WebSocketClosedError as e:
                    if self.failed_messages == 0:
                        self.logger.warning("Connection closed during async write_message")
                    self.failed_messages += 1
                    if self.ws_connection is not None:
                        self.ws_connection._abort()
                except Exception as e:
                    self.logger.warning("Exception during async write_message, stack at call:")
                    self.logger.warning("".join(cur_stack))
                    self.logger.warning(e, exc_info=True)
                    self.failed_messages += 1
                    if self.ws_connection is not None:
                        self.ws_connection._abort()

            # extreme back-compat try-except block, `f` should be None in
            # ancient tornado versions
            try:
                f.add_done_callback(after_write_callback)
            except:
                pass
            # true means that something was queued up to send, but it may be
            # async
            return True
        except tornado.websocket.WebSocketClosedError as e:
            if self.failed_messages == 0:
                self.logger.warning("Connection closed during write_message")
            self.failed_messages += 1
            if self.ws_connection is not None:
                self.ws_connection._abort()
        except:
            self.logger.warning("Exception trying to send message.", exc_info = True)
            self.failed_messages += 1
            if self.ws_connection is not None:
                self.ws_connection._abort()
            return False

    # send anything in the per-socket queue
    def flush_messages(self):
        # type: () -> bool
        if self.client_closed or len(self.message_queue) == 0:
            return False

        batch = ("{\"msgs\":["
            + ",".join(self.message_queue)
            + "]}")
        self.message_queue = [] # always empty the queue
        return self._send_raw_message(batch)

    # n.b. this looks a lot like superclass write_message, but has a static
    # type signature that is not compatible with it, so we do not override
    # that function.
    def append_message(self,
                       msg,      # type: str
                       send=True # type: bool
                       ):
        # type: (...) -> bool
        if self.client_closed:
            return False
        self.message_queue.append(msg)
        if send:
            return self.flush_messages()
        return False

    def send_message(self, msg, **data):
        # type: (str, Any) -> bool
        """Sends a JSON message to the client."""
        data["msg"] = msg
        return self.append_message(json_encode(data), True)

    def queue_message(self, msg, **data):
        # type: (str, Any) -> bool
        data["msg"] = msg
        return self.append_message(json_encode(data), False)

    def try_cleanup(self):
        global sockets
        # try to do what cleanup can be done on game end or close at the
        # current point in time.
        # If there is a running process, some cleanup will need to be
        # deferred, and if the client is not closed only the process will
        # be cleaned up.
        if not self.is_running():
            self.process = None
            if self.client_closed:
                self.clear_timeouts()
                if self in sockets:
                    sockets.remove(self)
            else:
                self.reset_idle_timeouts()

        if self.watched_game:
            self.watched_game.remove_watcher(self)

    def on_close(self):
        # at this point, self.client_closed is guaranteed to be true
        extra = []
        if self.is_running():
            # this will (sooner or later) trigger _on_crawl_end, which will
            # remove `self` from `sockets`
            # XX would it be better to handle on_close messaging at that point?
            self.process.stop()
        elif self in sockets and self.process:
            # buggy state (I think): `self.process` did not get set to
            # None, but process is not actually running.
            extra += ["stale process"]

        self.try_cleanup()

        if self.total_message_bytes == 0:
            comp_ratio = "N/A"
        else:
            comp_ratio = 100 - 100 * (self.compressed_bytes_sent + self.uncompressed_bytes_sent) / self.total_message_bytes
            comp_ratio = round(comp_ratio, 2)

        if self.failed_messages > 0 or self.failed_on_messages > 0:
            extra += ["%d (client) + %d (process) failed messages" % (self.failed_messages, self.failed_on_messages)]

        extra_msg = ""
        if len(extra):
            extra_msg = ", " + ", ".join(extra)

        self.logger.info("Socket closed. (%s sent, compression ratio %s%%%s)",
                         util.humanise_bytes(self.total_message_bytes),
                         comp_ratio,
                         extra_msg)



import collections
import logging
import os
import subprocess

import yaml
from tornado.escape import json_decode


# TODO: consider just reintegrating all this code into webtiles.config; its
# existence predates that module

# load game/template dicts from a single yaml file
def load_from_yaml(path):
    # some day, turn this into a data class
    result = dict(
        templates = collections.OrderedDict(),
        games = collections.OrderedDict(),
        source = path
        )

    try:
        with open(path) as f:
            yaml_text = f.read()
    except OSError as e:
        logging.warn("Couldn't read file %s: %s", path, e)
        yaml_text = None

    try:
        data = yaml.safe_load(yaml_text)
    except yaml.YAMLError as e:
        logging.warning("Failed to load games from %s, skipping (parse failure: %s).",
                        file_name, e)
        return result
    if not isinstance(data, dict) or ('games' not in data and 'templates' not in data):
        logging.warning("Failed to load games from %s, skipping (no 'games'/'templates' key).",
                        file_name)
        return result

    extra_keys = ",".join(key for key in data.keys() if key != 'games' and key != 'templates')
    if extra_keys:
        logging.warning(
            "Found extra top-level keys '%s' in %s, ignoring them"
            " (only 'games'/'templates' keys will be parsed).",
            extra_keys, file_name)


    for t in ('templates', 'games'):
        if t in data:
            for game in data[t]:
                if not 'id' in game:
                    logging.error("Definition missing id in %s! Skipping '%s'", t, repr(game))
                    continue
                if game['id'] in result[t]:
                    logging.warning(
                        "Duplicate id %s in %s (source %s), skipping.",
                        game['id'], t, path)
                    continue
                result[t][game['id']] = game
    return result


def merge_games(accum, new):
    import webtiles.config

    for t in ('templates', 'games'):
        if new[t] is None: # make it easy to clear via yml
            continue
        if len(new[t]):
            logging.info("Reading %d %s from %s", len(new[t]), t, new['source'])
        for k in new[t]:
            if k in accum[t]:
                logging.warning(
                    "Duplicate id %s in %s (source %s), skipping.",
                    k, t, new['source'])
                continue
            accum[t][k] = webtiles.config.GameConfig(new[t][k], game_id=k)
            if t == 'templates' and k != 'default' and webtiles.config.is_metatemplate(k):
                raise ValueError("Cannot redefine reserved template name `%s`" % k)
    return accum


def load_games(reloading=False):
    """
    Load game definitions from games.d/*.yaml and merge with existing config.

    This function will be called on startup and every time the webserver is sent
    a USR1 signal (eg `kill -USR1 <pid>`). This allows dynamic modification of
    the games list without interrupting active game sessions.

    The format of the source YAML files is: `games: [<game>, ...]`. Each game is
    a dictionary matching the format of entries in the `games` variable above,
    with an extra key `id` for the game's ID. Directory paths support %n as
    described above.

    Dynamic update notes and caveats:
    1. If you do modify a game entry, the changed settings only affect new game
       sessions (including new spectators). Existing sessions use the old config
       until they close.
    2. Some settings affect spectators. If you modify a running game's config,
       the mismatch of settings between the player and new spectators might
       cause spectating to fail until the player restarts with new settings.
    3. If something breaks on the reload, everything *should* roll back to the
       previous state. But it's going to be a lot safer not to rely on this.
    """
    import webtiles.config
    accum = dict(
        templates = collections.OrderedDict(),
        games = collections.OrderedDict(),
        )
    try:
        # first: load any games config info from the module
        # this will include any game defs in config.yml (or debug-config.yml)
        from_module = dict(
            templates = webtiles.config.get('templates'),
            games = webtiles.config.get('games'),
            source = "base config"
            )
        accum = merge_games(accum, from_module)

        # second: if circumstances warrant it, load games from yaml files in
        # the games.d subdirectory. By default, this is prevented if the
        # config module sets games.
        use_gamesd = webtiles.config.using_games_dir()
        if use_gamesd is None and not webtiles.config.get('games') or use_gamesd:
            base_path = os.path.join(webtiles.config.server_path, "games.d")
            if os.path.exists(base_path):
                for file_name in sorted(os.listdir(base_path)):
                    path = os.path.join(base_path, file_name)
                    if not path.endswith(".yaml"):
                        logging.info("Ignoring non-.yaml file %s", path)
                        continue
                    accum = merge_games(accum, load_from_yaml(path))
            else:
                logging.warn("Skipping non-existent YAML configuration directory: '%s'" % base_path)
        else:
            logging.warn("Skipping game definitions in `games.d/` based on config")

    except:
        if not reloading:
            raise # fail on initial load
        logging.error("Errors in game data, reverting to previous game configuration!", exc_info=True)
        return False

    old = dict(
        templates = webtiles.config.game_templates,
        games = webtiles.config.games)
    old_default = webtiles.config.get_template('default')

    # replace any existing config with what we have just loaded
    webtiles.config.game_templates = accum['templates']
    webtiles.config.games = accum['games']

    # do validation. We postpone this because success may rely on having an
    # updated `defaults` value in place.
    ok = True
    if ('default' in webtiles.config.game_templates and not
                            webtiles.config.define_default(
                                webtiles.config.game_templates.pop('default'))):
        ok = False

    for g in webtiles.config.game_templates.values():
        ok = g.validate() and ok

    if ok:
        # only validate games if templates are ok, since the main error case
        # leading to `not ok` here is a loop in templating (which will break
        # games that use that template)
        for g in webtiles.config.games.values():
            ok = g.validate_game() and ok

    if not ok:
        if not reloading:
            raise ValueError("Errors in game data!")
        logging.error("Errors in game data, reverting to previous game configuration!")
        webtiles.config.game_templates = old['templates']
        webtiles.config.metatemplates['default'] = old_default
        webtiles.config.games = old['games']
        return False

    # Done making changes. Print a diff message if reloading
    if reloading:
        # if we are reloading, print some helpful diff info
        for t in ('templates', 'games'):
            removed = [k for k in old[t] if k not in accum[t]]
            added = [k for k in accum[t] if k not in old[t]]
            # does the equality check work?
            updated = [k for k in accum[t] if k in old[t] and old[t][k] != accum[t][k]]
            if len(removed):
                logging.info("Removed %s: %s", t, ", ".join(removed))
            if len(added):
                logging.info("Added %s: %s", t, ", ".join(added))
            if len(updated):
                logging.info("Updated %s: %s", t, ", ".join(updated))
            else:
                logging.info("No updated %s", t)

    return True


def validate_game_dict(game):
    """Validate a game dictionary.

    Log warnings about issues and return validity.
    """
    if 'id' not in game:
        # Log the full game definition to help identify the game
        logging.warn("Missing 'id' from game definition %r", game)
        return False
    found_errors = False
    required = ('id', 'name', 'crawl_binary', 'rcfile_path', 'macro_path',
                'morgue_path', 'inprogress_path', 'ttyrec_path',
                'socket_path')
    optional = ('dir_path', 'cwd', 'morgue_url', 'milestone_path',
                'send_json_options', 'options', 'env', 'separator',
                'show_save_info', 'allowed_with_hold', 'version',
                'template', 'pre_options', 'client_path')
    # XX less ad hoc typing
    boolean = ('send_json_options', 'show_save_info', 'allowed_with_hold')
    string_array = ('options', 'pre_options')
    string_dict = ('env', )

    # values where %n can't be expanded
    # XX not sure this list gets everything
    username_invalid = ('pre_options', 'crawl_binary', 'socket_path')
    # XX should %v be validated here too? Currently handled in
    # GameConfig.validate_game

    allow_none = {'morgue_url'}
    for prop in required:
        if prop not in game:
            found_errors = True
            logging.warn("Missing required property '%s' in game '%s'",
                         prop, game['id'])
    for prop, value in game.items():
        expected = prop in required or prop in optional
        if not expected:
            # Don't count this as an error
            logging.warn("Unknown property '%s' in game '%s'",
                         prop, game['id'])
        if value is None and prop in allow_none:
            continue # short-circuit other checks

        if prop in boolean:
            if not isinstance(value, bool):
                found_errors = True
                logging.warn("Property '%s' value should be boolean in game '%s'",
                             prop, game['id'])
        elif prop in string_array:
            if not isinstance(value, list):
                found_errors = True
                logging.warn("Property '%s' value should be list of strings in game '%s'",
                             prop, game['id'])
                continue
            for item in value:
                if not isinstance(item, str):
                    found_errors = True
                    logging.warn(
                      "Item '%s' in property '%s' should be a string in game '%s'",
                      item, prop, game['id'])
            if prop in username_invalid and " ".join(value).find("%n") >= 0:
                found_errors = True
                logging.warn(
                    "'%%n' will not be expanded in %s (game %s): '%s'" % (
                                    rop, game['id'], repr(value)))
        elif prop in string_dict:
            if not isinstance(value, dict):
                found_errors = True
                logging.warn(
                  "Property '%s' value should be a map of string: string in game '%s'",
                  prop, game['id'])
                continue
            for item_key, item_value in value.items():
                if not isinstance(item_key, str):
                    found_errors = True
                    logging.warn(
                      "Item key '%s' in property '%s' should be a string in game '%s'",
                      item_key, prop, game['id'])
                if not isinstance(item_value, str):
                    found_errors = True
                    logging.warn("Item value '%s' of key '%s' in property '%s'"
                                 " should be a string in game '%s'",
                                 item_value, item_key, prop, game['id'])
                if prop in username_invalid and item_value.find("%n") >= 0:
                    found_errors = True
                    logging.warn(
                        "'%%n' will not be expanded in %s.%s (game %s): '%s'" % (
                                    prop, item_key, game['id'], repr(value)))
        else:
            # String property
            if not isinstance(value, str):
                found_errors = True
                logging.warn("Property '%s' value should be string in game '%s'",
                             prop, game['id'])
            if prop in username_invalid and value.find("%n") >= 0:
                found_errors = True
                logging.warn(
                    "'%%n' will not be expanded in %s (game %s): '%s'" % (
                                    prop, game['id'], repr(value)))
    return not found_errors


binaries = {}
def collect_game_modes():
    import webtiles.config
    # figure out what game modes are associated with which game in the config.
    # Basically: try to line up options in the game config with game types
    # reported by the binary. If the binary doesn't support `-gametypes-json`
    # then the type will be `None`. This is unfortunately fairly post-hoc, and
    # there would be much better ways of doing this if I weren't aiming for
    # backwards compatibility.
    #
    # This is very much a blocking call, especially with many binaries.
    # Changing crawl binaries in a way that affects this will require a restart
    # to get accurate save info.
    global binaries
    for g in webtiles.config.games:
        key = webtiles.config.games[g].get_binary_key()
        if not webtiles.config.games[g].get("show_save_info", False):
            binaries[key] = None
            continue
        if key in binaries and binaries[key]:
            continue
        call = webtiles.config.games[g].get_call_base()

        # "dummy" is here for the sake of the dgamelaunch-config launcher
        # scripts, which choke badly if there is no second argument. The actual
        # crawl binary just ignores it. (TODO: this is not an ideal thing about
        # these wrapper scripts...)
        call += ["-gametypes-json", "dummy"]
        try:
            m_json = subprocess.check_output(call, stderr=subprocess.STDOUT)
            binaries[key] = json_decode(m_json)
        except subprocess.CalledProcessError:  # return value 1
            binaries[key] = None
        except ValueError:  # JSON decoding issue?
            logging.warn("JSON error with output '%s'", repr(m_json))
            binaries[key] = None

    # Example output of this call:
    # {"normal":"","tutorial":"-tutorial","arena":"-arena","sprint":"-sprint","seeded":"-seed"}
    # we want to take the options specified in this map and reconstruct the
    # game modes as defined in the game configs.

    game_modes = {}
    for g in webtiles.config.games:
        key = webtiles.config.games[g].get_binary_key()
        if binaries[key] is None:
            # this binary either has no enabled save infos, opr does not
            # support -gamemode-json, so we know nothing about its game modes.
            # (Pre-0.24 dcss.)
            # XX should we always try to collect game modes?
            game_modes[g] = None
            continue

        # A gametype may specify "" as the options string (e.g. "normal"). We
        # can't look directly for this, since there could be other options.
        # So first find the relevant mode if any, and then use it as a fallback
        # for any calls where no modes are found.
        no_options_mode = None
        for m in binaries[key]:
            if binaries[key][m] == "":
                no_options_mode = m
                break

        mode_found = False
        for mode in binaries[key]:
            for opt in webtiles.config.game_param(g, "options", default=[]):
                if opt == binaries[key][mode]:
                    game_modes[g] = mode
                    mode_found = True
                    break
            if mode_found:
                break
        if not mode_found and no_options_mode is not None:
            game_modes[g] = no_options_mode

    return game_modes



import datetime

import pytest

from webtiles import auth

try:
    import mock
except ImportError:
    from unittest import mock


class Test_purge_login_tokens:

    @mock.patch('webtiles.auth.datetime')
    def test_not_yet_expired_tokens_are_kept(self, stub_datetime):
        auth.login_tokens = {
            ('token', 'username'): datetime.datetime.max,
        }
        stub_datetime.datetime.now.return_value = datetime.datetime.min

        auth.purge_login_tokens()

        assert auth.login_tokens

    @mock.patch('webtiles.auth.datetime')
    def test_expired_tokens_are_discarded(self, stub_datetime):
        auth.login_tokens = {
            ('token', 'username'): datetime.datetime.min,
        }
        stub_datetime.datetime.now.return_value = datetime.datetime.max

        auth.purge_login_tokens()

        assert not auth.login_tokens


class Test_log_in_as_user:

    @mock.patch('webtiles.auth.datetime')
    def test_cookies_have_correct_lifetime(self, stub_datetime):
        auth.config.set('login_token_lifetime', 42)
        stub_datetime.datetime.now.return_value = datetime.datetime.min
        stub_datetime.timedelta = datetime.timedelta
        stub_request = mock.Mock()

        auth.log_in_as_user(stub_request, 'user')

        expires = list(auth.login_tokens.values())[0]
        lifetime = expires - datetime.datetime.min

        assert lifetime == datetime.timedelta(42)

    def test_returning_user_can_log_in_automatically(self):
        stub_request = mock.Mock()
        username = 'user'

        cookie = auth.log_in_as_user(stub_request, username)

        assert auth.check_login_cookie(cookie) == (username, True)

    def test_log_out_removes_token(self):
        stub_request = mock.Mock()

        cookie = auth.log_in_as_user(stub_request, 'user')
        num_tokens = len(auth.login_tokens)
        auth.forget_login_cookie(cookie)

        assert len(auth.login_tokens) == num_tokens - 1

    def test_repeated_log_out_is_idempotent(self):
        stub_request = mock.Mock()

        cookie = auth.log_in_as_user(stub_request, 'user')
        auth.forget_login_cookie(cookie)
        login_tokens = auth.login_tokens
        auth.forget_login_cookie(cookie)  # Repeat call should do nothing

        assert login_tokens == auth.login_tokens


class Test__parse_login_cookie:

    @pytest.mark.parametrize("cookie", [
        "",
        " ",
        "    ",
        " asdf",
        "asdf",
        "asdf ",
        "asdf asdf",
        "asdf asdf asdf",
    ])
    def test_returns_none_for_wrong_format(self, cookie):
        username, token = auth._parse_login_cookie(cookie)
        assert token is None

    @pytest.mark.parametrize("cookie", [
        "abc%20123",
    ])
    def test_returns_username_and_token(self, cookie):
        username, token = auth._parse_login_cookie(cookie)
        assert username and token



import argparse
import asyncio
import errno
import functools
import logging
import logging.handlers
import os
import os.path
import sys
import signal
import time

import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.template
import tornado.web
from tornado.ioloop import IOLoop
from tornado.netutil import bind_sockets
# do not add tornado.platform here without changing do_chroot()

import webtiles
from webtiles import auth, load_games, process_handler, userdb, config
from webtiles import game_data_handler, util, ws_handler, status


servers = None
https_port = None


# port should already be formatted with a ":"
class HTTPSRedirectHandler(tornado.web.RequestHandler):
    def get(self):
        global https_port
        if https_port is None:
            # this will probably break unless 80/443 are in use
            https_port = ""
        self.redirect(f"https://{self.request.host_name}{https_port}{self.request.uri}", permanent=True)


class MainHandler(tornado.web.RequestHandler):
    # async def _execute(self, transforms, *args, **kwargs):
    #     await tornado.web.RequestHandler._execute(self, transforms, *args, **kwargs)

    def get(self):
        host = self.request.host
        if self.request.protocol == "https" or self.request.headers.get("x-forwarded-proto") == "https":
            protocol = "wss://"
        else:
            protocol = "ws://"

        recovery_token = None
        recovery_token_error = None

        recovery_token = self.get_argument("ResetToken",None)
        if recovery_token:
            recovery_token_error = userdb.find_recovery_token(recovery_token)[2]
            if recovery_token_error:
                logging.warning("Recovery token error from %s", self.request.remote_ip)

        with util.SlowWarning("Slow IO: render client.html"):
            self.render("client.html", socket_server = protocol + host + "/socket",
                    game_version = _crawl_version,
                    username = None,
                    config = config,
                    reset_token = recovery_token, reset_token_error = recovery_token_error)

class NoCacheHandler(tornado.web.StaticFileHandler):
    def set_extra_headers(self, path):
        self.set_header("Cache-Control", "no-cache, no-store, must-revalidate")
        self.set_header("Pragma", "no-cache")
        self.set_header("Expires", "0")


_crawl_version = "unknown"

def load_version():
    global _crawl_version
    try:
        # this is the "bad" way to do this. However, the supposedly good ways
        # to do this are all insanely complicated in ways that are irrelevant
        # to webtiles.
        with open(os.path.join(os.path.dirname(__file__), 'version.txt')) as f:
            _crawl_version = f.readline().strip()
    except:
        _crawl_version = "unknown"


def version_data():
    v = sys.version_info
    supported = 8 # minimum version, see https://devguide.python.org/versions/
    return dict(
        webtiles=_crawl_version,
        tornado=tornado.version,
        python="%d.%d.%d" % (v[0], v[1], v[2]),
        python_supported=v[0] >= 3 and v[1] >= supported)


def version():
    vdata = version_data()
    # TODO convert to f string some day...
    result = "Webtiles (%s) running with Tornado %s and Python %s" % (
        vdata["webtiles"],
        vdata["tornado"],
        vdata["python"])
    if not vdata['python_supported']:
        result = result + "\nUnsupported Python version! Please use a supported version: https://devguide.python.org/versions/"
    return result

def err_exit(errmsg, exc_info=False):
    if exc_info or config.get('logging_config').get('filename'):
        # don't print duplicate messages on stdout
        logging.error(errmsg, exc_info=exc_info)
    sys.exit(errmsg)

def daemonize():
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)
    except OSError as e:
        err_exit("Fork #1 failed! (%s)" % (e.strerror))

    os.setsid()

    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)
    except OSError as e:
        err_exit("Fork #2 failed! (%s)" % e.strerror)

    with open("/dev/null", "r") as f:
        os.dup2(f.fileno(), sys.stdin.fileno())
    with open("/dev/null", "w") as f:
        os.dup2(f.fileno(), sys.stdout.fileno())
        os.dup2(f.fileno(), sys.stderr.fileno())

def write_pidfile():
    pidfile = config.get('pidfile')
    if not pidfile:
        return
    if os.path.exists(pidfile):
        try:
            with open(pidfile) as f:
                pid = int(f.read())
        except ValueError:
            err_exit("PIDfile %s contains non-numeric value" % pidfile)
        try:
            os.kill(pid, 0)
        except OSError as why:
            if why.errno == errno.ESRCH:
                # The pid doesn't exist.
                logging.warn("Removing stale pidfile %s" % pidfile)
                os.remove(pidfile)
            else:
                err_exit("Can't check status of PID %s from pidfile %s: %s" %
                         (pid, pidfile, why.strerror))
        else:
            err_exit("Another Webtiles server is running, PID %s\n" % pid)

    with open(pidfile, "w") as f:
        f.write(str(os.getpid()))

def remove_pidfile():
    pidfile = config.get('pidfile')
    if not pidfile:
        return

    try:
        os.remove(pidfile)
    except OSError as e:
        if e.errno == errno.EACCES or e.errno == errno.EPERM:
            logging.warn("No permission to delete pidfile!")
        else:
            logging.error("Failed to delete pidfile!")
    except:
        logging.error("Failed to delete pidfile!")

def shed_privileges():
    if config.get('gid') is not None:
        os.setgid(config.get('gid'))
    if config.get('uid') is not None:
        os.setuid(config.get('uid'))


def all_tasks_compat():
    try:
        # only present since py37; maybe safe to assume?
        return asyncio.all_tasks()
    except AttributeError:
        # removed in py39
        return asyncio.Task.all_tasks()


def stop_everything(shutdown_event):
    # prevent reload signals from interacting with shutdown
    if config.get('hup_reloads_config'):
        asyncio.get_event_loop().remove_signal_handler(signal.SIGHUP)
    asyncio.get_event_loop().remove_signal_handler(signal.SIGUSR1)

    global servers
    # shut down servers first -- stop accepting new connections
    for server in servers:
        server.stop()
    # shut down ongoing games
    # XX any errors in this code will prevent shutdown, do something more?
    ws_handler.shutdown()

    # an open question -- are there cases where even stronger measures might be
    # needed for stuck processes? task.cancel() handled everything I could think
    # of locally.
    def do_stop(tries=0):
        if len(ws_handler.sockets) == 0:
            shutdown_event.set()
        elif tries > 60: # 30s
            # try something a bit stronger
            logging.error("Stop failed after 30s, cancelling remaining tasks.")
            logging.error("Remaining sockets: %s", ws_handler.describe_sockets(names=True))
            for task in all_tasks_compat():
                task.cancel()
        else:
            IOLoop.current().add_timeout(time.time() + 0.5,
                functools.partial(do_stop, tries=tries + 1))

    do_stop()

def stop_handler(signum, shutdown_event):
    logging.info("Received signal %i, beginning shutdown.", signum)
    IOLoop.current().add_callback(functools.partial(stop_everything, shutdown_event))


def reload_config_handler(signum):
    logging.info("Received signal %i, reloading all config data.", signum)
    IOLoop.current().add_callback(config.reload)


def reload_games_handler(signum):
    logging.info("Received signal %i, reloading game config data.", signum)
    IOLoop.current().add_callback_from_signal(config.load_game_data)


def bind_server_sockets():
    nonsecure = []
    secure = []
    if config.get('bind_nonsecure'):
        if config.get('bind_pairs'):
            listens = config.get('bind_pairs')
        else:
            listens = ( (config.get('bind_address'), config.get('bind_port')), )
        for (addr, port) in listens:
            logging.info("Listening on http://%s:%d" % (addr, port))
            nonsecure.append(bind_sockets(port, addr))

    if config.get('ssl_options'):
        if config.get('ssl_bind_pairs'):
            listens = config.get('ssl_bind_pairs')
        else:
            listens = ( (config.get('ssl_address'), config.get('ssl_port')), )
        global https_port
        for (addr, port) in listens:
            logging.info("Listening on https://%s:%d" % (addr, port))
            if https_port is None:
                https_port = f":{port}"
            secure.append(bind_sockets(port, addr))
    return nonsecure, secure

def bind_server(nonsecure_sockets, secure_sockets):
    settings = {
        "static_path": config.get('static_path'),
        "template_loader": util.DynamicTemplateLoader.get(config.get('template_path')),
        "debug": bool(config.get('development_mode', False)),
        }

    if config.get('no_cache', False):
        settings["static_handler_class"] = NoCacheHandler

    handlers = [
            (r"/", MainHandler),
            (r"/socket", ws_handler.CrawlWebSocket),
            (r"/gamedata/([0-9a-f]*\/.*)", game_data_handler.GameDataHandler),
            (r"/status/lobby/", status.LobbyHandler),
            (r"/status/version/", status.VersionHandler),
            ]

    application = tornado.web.Application(handlers,
                                    gzip=config.get('use_gzip'), **settings)

    kwargs = {}
    if config.get('http_connection_timeout') is not None:
        kwargs["idle_connection_timeout"] = config.get('http_connection_timeout')

    # TODO: the logic looks odd here, as it is set to None in the default
    # config.py. But I'm not really sure how this is used so I don't want to
    # mess with it...
    if config.get('http_xheaders', False):
        kwargs["xheaders"] = config.get('http_xheaders')

    global servers
    servers = []

    if nonsecure_sockets:
        if config.get('bind_nonsecure') == "redirect":
            http_app = tornado.web.Application([(r"/", HTTPSRedirectHandler)])
        else:
            http_app = application
        server = tornado.httpserver.HTTPServer(http_app, **kwargs)
        for s in nonsecure_sockets:
            server.add_sockets(s)
        servers.append(server)

    if secure_sockets:
        # TODO: allow different ssl_options per bind pair?
        server = tornado.httpserver.HTTPServer(application,
                            ssl_options=config.get('ssl_options'), **kwargs)
        for s in secure_sockets:
            server.add_sockets(s)
        servers.append(server)

    if not servers:
        # config validation should preempt this, but it's here for robustness
        raise ValueError("No ports succesfully configured!")
    return servers


def init_logging(logging_config):
    filename = logging_config.get("filename")
    if filename:
        max_bytes = logging_config.get("max_bytes", 10*1000*1000)
        backup_count = logging_config.get("backup_count", 5)
        hdlr = logging.handlers.RotatingFileHandler(
            filename, maxBytes=max_bytes, backupCount=backup_count)
    else:
        hdlr = logging.StreamHandler(None)
    fs = logging_config.get("format", "%(levelname)s:%(name)s:%(message)s")
    dfs = logging_config.get("datefmt", None)
    fmt = logging.Formatter(fs, dfs)
    hdlr.setFormatter(fmt)
    logging.getLogger().addHandler(hdlr)
    level = logging_config.get("level")
    if level is not None:
        logging.getLogger().setLevel(level)
    if tornado.version_info[0] >= 3:
        # hide regular successful access messages, e.g. `200 GET` messages
        # messages. 404s are still shown. TODO: would there be demand for
        # sending this to its own file in a configurable way?
        logging.getLogger("tornado.access").setLevel(logging.WARNING)
    else:
        # the tornado version is ancient enough that it doesn't have its own
        # logging streams; this filter suppresses any logging done from the
        # `web` module at level INFO.
        logging.getLogger().addFilter(util.TornadoFilter())
    logging.addLevelName(logging.DEBUG, "DEBG")
    logging.addLevelName(logging.WARNING, "WARN")


def parse_args_main():
    parser = argparse.ArgumentParser(
        description='Dungeon Crawl webtiles server',
        epilog='Command line options will override config settings. See wtutil.py for database commands.')
    parser.add_argument('-p', '--port', type=int, help='A port to bind; disables SSL.')
    parser.add_argument('--ssl-port', type=int, help='An SSL port to bind. Requires configured `ssl_options`.')
    parser.add_argument('--logfile',
                        help='A logfile to write to; use "-" for stdout.')
    parser.add_argument('--daemon', action='store_true', default=None,
                        help='Daemonize after start.')
    parser.add_argument('-n', '--no-daemon', action='store_false', default=None,
                        dest='daemon',
                        help='Do not daemonize after start.')
    parser.add_argument('--no-pidfile', dest='pidfile', action='store_false',
                        default=None, help='Do not use a PID-file.')

    # live debug mode is intended to be able to run (more or less) safely with
    # a concurrent real webtiles server. However, still be careful with this...
    parser.add_argument('--live-debug', action='store_true',
        help=('Debug mode for server admins. Will use a separate directory for sockets. '
              'Entails --no-pidfile, --no-daemon, --logfile -, watch_socket_dirs=False. '
              '(Further command line options can override these.)'))
    result = parser.parse_args()
    if result.live_debug:
        if not result.logfile:
            result.logfile = '-'
        if result.daemon is None:
            result.daemon = False
        result.pidfile = False
        config.set('live_debug', True)
    return result


# override config with any arguments supplied on the command line
def export_args_to_config(args):
    if config.get('live_debug'):
        config.server_config._load_override_file(os.path.join(
                            config.get("server_path", ""), "debug-config.yml"))
        config.do_early_logging() # sigh
    if args.port:
        if args.ssl_port:
            err_exit("Can't combine --port and --ssl-port.")
        config.set('bind_nonsecure', True)
        config.set('bind_address', "")  # TODO: ??
        config.set('bind_port', args.port)
        if config.get('bind_pairs') is not None:
            config.pop('bind_pairs')
        logging.info("Using command-line supplied port: %d", args.port)
        if config.get('ssl_options'):
            logging.info("    (Overrides config-specified SSL settings.)")
            config.set('ssl_options', None)
    elif args.ssl_port:
        config.set('bind_nonsecure', False)
        config.set('ssl_bind_pairs', (('', args.ssl_port),))
        if not config.get('ssl_options'):
            err_exit("--ssl-port option requires configured `ssl_options`")
        logging.info("Using command-line supplied ssl port: %d", args.ssl_port)
    if args.daemon is not None:
        logging.info("Command line override for daemonize: %r", args.daemon)
        config.set('daemon', args.daemon)
    if args.pidfile is not None:
        if not args.pidfile:
            if config.get('pidfile'):
                logging.info("Command line overrides config-specified PID file!")
            config.set('pidfile', None)


def reset_token_commands(args):
    if args.clear_reset:
        username = args.clear_reset
    else:
        username = args.reset

    user_info = userdb.get_user_info(username)
    if not user_info:
        err_exit("Reset/clear password failed; invalid user: %s" % username)

    username = user_info.username # canonicalize

    # don't crash on the default config
    if config.get('lobby_url') is None:
        config.set('lobby_url', "[insert lobby url here]")

    if args.clear_reset:
        ok, msg = userdb.clear_password_token(username)
        if not ok:
            err_exit("Error clearing password reset token for %s: %s" % (username, msg))
        else:
            print("Password reset token cleared for account '%s'." % username)
            return True
    else:
        ok, msg = userdb.generate_forgot_password(username)
        if not ok:
            err_exit("Error generating password reset token for %s: %s" % (username, msg))
        else:
            if not user_info.email:
                logging.warning("No email set for account '%s', use caution!" % username)
            print("Setting a password reset token on account '%s'." % username)
            print("Email: %s\nMessage body to send to user:\n%s\n" % (user_info.email, msg))
            return True
    return False


def mode_to_flag(mode):
    if mode == "admin":
        return userdb.DGLACCT_ADMIN
    elif mode == "ban":
        return userdb.DGLACCT_LOGIN_LOCK
    elif mode == "hold":
        return userdb.DGLACCT_ACCOUNT_HOLD
    elif mode == "wizard":
        if not config.get('wizard_accounts'):
            logging.warning("Wizard accounts not enabled on this server, do you want `admin`?")
        return userdb.DGLACCT_WIZARD
    elif mode == "bot":
        if not config.get('bot_accounts'):
            logging.warning("Bot accounts are not enabled on this server...")
        return userdb.DGLACCT_BOT
    return 0


def show_flags(username):
    r = userdb.get_user_info(username)
    if not r:
        err_exit("Unknown user '%s'!" % username)
    # XX would be nice to normalize username
    if not r.flags:
        print("User '%s' (id %d) has no flags set." % (r.username, r.id))
    else:
        print("Flags for '%s' (id %d): %s" % (r.username, r.id, userdb.flag_description(r.flags)))
    return True


def flag_commands(args):
    if args.show:
        return show_flags(args.show)

    # everything below here needs an explicit flag
    if not args.flag:
        print("Missing flag to query!", file=sys.stderr)
        return False
    flag = mode_to_flag(args.flag)
    if not flag:
        print("Unknown flag '%s'" % args.list, file=sys.stderr)
        return False

    if args.list:
        l = userdb.get_users_by_flag(flag)
        if not l:
            print("No matching users.")
        else:
            print("Users with flag '%s': %s" %
                (userdb.flag_description(flag), ", ".join([u.username for u in l])))
        return True
    elif args.set:
        r = userdb.set_flags(args.set, flag, mask=flag)
        if r:
            err_exit("Error when setting flag %s for '%s': %s" % (args.set, arg.flag, r))
        else:
            print("Flag set.")
            show_flags(args.set)
        return True
    elif args.clear:
        r = userdb.set_flags(args.clear, 0, mask=flag)
        if r:
            err_exit("Error when clearing flag %s for '%s': %s" % (args.clear, arg.flag, r))
        else:
            print("Flag cleared.")
            show_flags(args.clear)
        return True
    return False


def ban_commands(args):
    if args.check_config_bans or args.run_config_bans:
        # potentially very heavy commands on old servers...
        all_users = userdb.get_all_users()
        affected = [n.username for n in all_users if not config.check_name(n.username)]
        if not affected:
            print("No affected users.")
            return True
        if args.check_config_bans:
            print("%d existing user(s) have banned names:" % len(affected))
        else:
            print("Banning %d user(s)!" % len(affected))
            for n in affected:
                userdb.set_ban(n, True)
        if affected:
            print("    Affected users: %s" % ', '.join(affected))
        return True
    elif args.list or args.list_holds:
        banned, held = userdb.get_bans()
        if args.list:
            if len(banned):
                print("Banned users: " + ", ".join(banned))
            else:
                print("No banned users!")
        if len(held):
            print("Account holds: " + ", ".join(held))
        else:
            print("No account holds!")
        return True
    elif args.clear_holds:
        banned, held = userdb.get_bans()
        if len(held):
            cleared = []
            for u in held:
                err = userdb.set_account_hold(u, False)
                if err:
                    err_exit("Error when clearing hold for '%s': %s" % (u, err))
                else:
                    cleared.append(u)
            if len(cleared):
                print("Holds cleared for: " + ", ".join(cleared))
            else:
                print("No holds cleared.")
        else:
            print("No holds to clear.")
        return True
    elif args.add:
        err = userdb.set_ban(args.add, True)
        if err:
            err_exit("Error when trying to ban user '%s': %s" % (args.add, err))
        else:
            print("'%s' is now banned." % args.add)
        return True
    elif args.hold:
        err = userdb.set_account_hold(args.hold, True)
        if err:
            err_exit("Error when trying to set hold for user '%s': %s" % (args.hold, err))
        else:
            print("Account hold set for '%s'." % args.hold)
        return True
    elif args.clear:
        err1 = userdb.set_account_hold(args.clear, False)
        if err1:
            err2 = userdb.set_ban(args.clear, False)
            if err2:
                err_exit("Error when trying to clear hold for user '%s': %s" % (args.clear, err2))
            else:
                print("'%s' is no longer banned." % args.clear)
        else:
            print("Account hold cleared for '%s'." % args.clear)
        return True
    return False


def parse_args_util():
    parser = argparse.ArgumentParser(
        description='Dungeon Crawl webtiles utilities.',
        epilog='Command line options will override config settings.')
    parser.add_argument('--logfile',
        default='-',
        help='A logfile to write to; use "-" for stdout (the default, overrides config).')
    subparsers = parser.add_subparsers(help='Mode', dest='mode')
    subparsers.required = True
    parser_pw = subparsers.add_parser('password',
        help="Set and clear password reset tokens.",
        description="Set and clear password reset tokens.")
    parser_pw.add_argument('--reset', type=str,
        help='Generate a password reset token for the specified username.')
    parser_pw.add_argument('--clear-reset', type=str,
        help='Clear any password reset tokens for the specified username.')
    parser_ban = subparsers.add_parser('ban',
        help="Set and clear bans and account holds by username.",
        description="Set and clear bans and account holds by username.")
    parser_ban.add_argument('--add', type=str, help='Set an account ban for a user. Replaces any account holds.')
    parser_ban.add_argument('--hold', type=str, help='Set an account hold for a user. Replaces any ban flags.')
    parser_ban.add_argument('--clear', type=str, help='Clear a ban/hold for a user.')
    parser_ban.add_argument('--run-config-bans', action='store_true',
        help='Ban any existing usernames disallowed by server config banned name options')
    parser_ban.add_argument('--check-config-bans', action='store_true',
        help='List any existing usernames that would be banned by --run-config-bans')
    parser_ban.add_argument('--list', action='store_true',
            help='List current banned users and account holds.')
    parser_ban.add_argument('--list-holds', action='store_true',
            help='List current account holds.')
    parser_ban.add_argument ('--clear-holds', action='store_true',
            help='Clear all current account holds')

    parser_flag = subparsers.add_parser('flag',
        help="Low level interface to query and set userdb flags.",
        description="Low-level interface to query and set userdb flags. For managing bans/holds, the 'ban' command is recommended instead.")
    parser_flag.add_argument('flag', type=str, nargs='?', default="", help='a flag to query with (ban, hold, admin, wizard, bot)')
    parser_flag.add_argument('--list', action='store_true',
        help='List any users with the indicated flag')
    parser_flag.add_argument('--set', type=str, help='Set a flag on a specified user.')
    parser_flag.add_argument('--clear', type=str, help='Clear a flag on a specified user.')
    parser_flag.add_argument('--show', type=str, help='Show flags on a specified user (does not require [flag]).')

    result = parser.parse_args()
    help_fun = parser.print_help
    if result.mode == "password":
        help_fun = parser_pw.print_help
    elif result.mode == "ban":
        help_fun = parser_ban.print_help
    elif result.mode == "flag":
        help_fun = parser_flag.print_help

    if not result.logfile:
        result.logfile = '-'
    result.daemon = False
    result.pidfile = False
    return result, help_fun


def do_chroot():
    if config.get('chroot'):
        os.chroot(config.get('chroot'))
        try:
            # try to fail early, with an informative message, if this is not
            # going to work
            # the choice of tornado.platform is a bit heuristic, but it is
            # currently where the webserver seems to fail on import after a chroot.
            # If it is ever imported at the top of this file, something else would
            # be needed.
            import tornado.platform
        except:
            # no logging available yet
            print("Error: can't import `tornado.platform` in chroot. Did you copy"
                " the python library for this version of python into the chroot?",
                file=sys.stderr)
            raise


def run_util():
    args, help_fun = parse_args_util()
    do_chroot()

    if config.source_file is None:
        sys.exit("No configuration provided!")

    try:
        config.validate()
    except:
        err_exit("Errors in config. Exiting.", exc_info=True)

    # duplicate some minimal setup needed for this to work. Note that unlike
    # the main server.py call, this arg parser sets '-' as the default for
    # the log, overriding config.
    if not args.logfile or args.logfile == "-":
        config.get('logging_config').pop('filename', None)
    else:
        config.get('logging_config')['filename'] = args.logfile

    init_logging(config.get('logging_config'))

    userdb.init_db_connections()

    mode_fun = None

    if args.mode == "password":
        mode_fun = reset_token_commands
    elif args.mode == "ban":
        mode_fun = ban_commands
    elif args.mode == "flag":
        mode_fun = flag_commands

    if not mode_fun or not mode_fun(args):
        help_fun()
        sys.exit(1)


def init_signals(shutdown_event):
    # if we can ensure >=py37 (probably?) this should be get_running_loop
    loop = asyncio.get_event_loop()
    stop_sigs = [signal.SIGTERM, signal.SIGINT]
    reload_sigs = []
    reload_game_sigs = [signal.SIGUSR1]

    if config.get('hup_reloads_config'):
        reload_sigs.append(signal.SIGHUP)
    else:
        stop_sigs.append(signal.SIGHUP)

    for s in stop_sigs:
        loop.add_signal_handler(s, functools.partial(stop_handler, s, shutdown_event))
    for s in reload_sigs:
        loop.add_signal_handler(s, functools.partial(reload_config_handler, s))
    for s in reload_game_sigs:
        loop.add_signal_handler(s, functools.partial(reload_games_handler, s))


async def async_run_server(nonsecure_sockets, secure_sockets):
    # is this ever set to False by anyone in practice?
    dgl_mode = config.get('dgl_mode')
    # XX possibly some of this should move into async_run
    if dgl_mode:
        ws_handler.status_file_timeout() # note: tornado coroutine
        auth.purge_login_tokens_timeout()
        ws_handler.start_reading_milestones()

        if config.get('watch_socket_dirs'):
            process_handler.watch_socket_dirs()

    # set up various timeout loops
    ws_handler.do_periodic_lobby_updates()
    webtiles.config.init_config_timeouts()

    bind_server(nonsecure_sockets, secure_sockets)

    shutdown_event = tornado.locks.Event()
    init_signals(shutdown_event)

    logging.info("DCSS Webtiles server started! (PID: %s)" % os.getpid())
    logging.info(version())
    await shutdown_event.wait()


# before running, this needs to have its config source set up. See
# ../server.py in the official repository for an example.
def run():
    args = parse_args_main()
    do_chroot()

    if config.source_file is None:
        # we could try to automatically figure this out from server_path, if
        # it is set?
        sys.exit("No configuration provided!")

    # do this here so it can happen before logging init
    if args.logfile:
        if args.logfile == "-":
            config.get('logging_config').pop('filename', None)
            args.logfile = "<stdout>"  # make the log message easier to read
        else:
            config.get('logging_config')['filename'] = args.logfile

    init_logging(config.get('logging_config'))
    try:
        logging.info("Loaded server configuration from: %s", config.source_file)
        config.do_early_logging()

        if config.get('live_debug'):
            logging.info("Starting in live-debug mode.")
            config.set('watch_socket_dirs', False)

        if args.logfile:
            logging.info("Using command-line supplied logfile: '%s'", args.logfile)

        export_args_to_config(args)

        try:
            config.load_game_data()
            config.validate()
        except:
            err_exit("Errors in game data. Exiting.", exc_info=True)

        if config.get('daemon', False):
            daemonize()

        if config.get('umask') is not None:
            os.umask(config.get('umask'))
        # bind sockets and shed privileges before starting up the ioloop
        nonsecure_sockets, secure_sockets = bind_server_sockets()
        # note -- shed_privileges cannot move later, or various files end up
        # owned by root, breaking dgl-config installs! Particularly pidfile,
        # but it's not the only one.
        shed_privileges()

    except SystemExit: # err_exit in the try blocks
        # logging already done, hopefully
        raise
    except:
        err_exit("Server startup failed!", exc_info=True)

    try:
        write_pidfile()
        userdb.init_db_connections()
        try:
            # finally -- start things up for real
            asyncio.run(async_run_server(nonsecure_sockets, secure_sockets))
            logging.info("Bye!")
        except asyncio.exceptions.CancelledError:
            # triggered by the cancel case in stop_everything
            err_exit("Normal server stop failed, some tasks were force-cancelled!")
        except SystemExit:
            raise
        except:
            err_exit("Server exited with error!", exc_info=True)

    except SystemExit:
        raise
    except:
        err_exit("Server startup failed!", exc_info=True)
    finally:
        # warning: need to be careful what appears in this finally block, since
        # it may be called by child processes on fork in terminal.py in the
        # event of rare bad timing or bugs. (So any global state that may be
        # used here should be reset in the child process...)
        remove_pidfile()


load_version()

# TODO: it might be nice to simply make this module runnable, but that would
# need some way of specifying the config source and `config.server_path`.



#!/usr/bin/env python3

"""
Works with both Python 2 & 3.
"""

import subprocess
import sys

if len(sys.argv) < 3:
    sys.stderr.write('Error: missing arguments\n')
    sys.exit(1)

cxx = sys.argv[1]
includesStr = sys.argv[2]

includes = includesStr.split(';')
includes = [x.strip() for x in includes]
includes = [x for x in includes if x]

arguments = '"' + cxx + '"'
if len(includes) != 0:
    arguments += ' /I "'
    arguments += '" /I "'.join(includes)
    arguments += '"'

python = sys.executable
command = [python, 'util/configure.py', arguments]
result = subprocess.call(command)
if(result != 0):
    sys.exit(result)
